/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												INCLUDES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#include "Actor.h"
#include "Globals.h"
#include "Optics.h"
#include "MessageDispatcher.h"
#include "MemoryPool.h"
#include "Printing.h"


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											 CLASS'S MACROS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

enum CollidingObjectIndexes{
	eXAxis = 0,
	eYAxis,
	eZAxis,
	eLastCollidingObject,
};

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


__CLASS_DEFINITION(Actor, InGameEntity); 

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

		
//};
/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// always call these to macros next to each other
__CLASS_NEW_DEFINITION(Actor, __PARAMETERS(ActorDefinition* actorDefinition, int ID))
__CLASS_NEW_END(Actor, __ARGUMENTS(this, actorDefinition, ID));

// Actor.c 
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's conctructor
void Actor_constructor(Actor this, ActorDefinition* actorDefinition, int ID){

	// construct base object
	__CONSTRUCT_BASE(InGameEntity, __ARGUMENTS(this, &actorDefinition->inGameEntityDefinition, ID));
	
	// create the animated sprite
	this->sprite = (Sprite)__NEW(AnimatedSprite, 
					__ARGUMENTS((void*)this, 
					actorDefinition->animationDescription->numberOfFrames,
					&actorDefinition->inGameEntityDefinition.entityDefinition.spriteDefinition
					));
	
	// write the sprite to graphic memory
	__VIRTUAL_CALL(void, Sprite, write, this->sprite);
	
	// save ROM definition
	this->actorDefinition = actorDefinition;
	
	// construct the game state machine
	this->stateMachine = __NEW(StateMachine, __ARGUMENTS(this));
	
	//set the direction
	this->direction.x = __RIGHT;
	this->previousDirection.x = __LEFT;
	this->direction.y = __DOWN;
	this->direction.z = __FAR;
	
	//state ALIVE for initial update
	this->inGameState = kLoaded;
	
	// reset action timers
	this->timeX = 0;
	this->timeY = 0;
	this->timeZ = 0;
	
	//pointing to right
	this->direction.alpha.cos = 0;
	this->direction.betha.cos = 0;
	this->direction.tetha.cos = 0;

	// clear movement type
	this->moveType = 0;
	
	this->velocity.x = 0;
	this->velocity.y = 0;
	this->velocity.z = 0;
	
	this->acceleration.x = 0;
	this->acceleration.y = 0;
	this->acceleration.z = 0;
	
	
	this->movementState.x = __PASSIVE;
	this->movementState.y = __PASSIVE;
	this->movementState.z = __PASSIVE;
	
	this->velocity.x = 0;
	this->velocity.y = 0;
	this->velocity.z = 0;
	
	Actor_calculateAnglesX(this);
	Actor_calculateAnglesY(this);
	Actor_calculateAnglesZ(this);
	
	this->objectBelow = NULL;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
void Actor_destructor(Actor this){

	
	// destroy state machine
	__DELETE(this->stateMachine);

	// destroy the super object
	__DEALLOCATE(InGameEntity);
	

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set character's angle
void Actor_setAngle(Actor this, Angle angle){
	//this->direction.alpha.degree=angle.degree;
	//this->direction.betha.degree=angle.degree;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// calculate character's direction angle
void Actor_calculateAnglesX(Actor this){
	
	this->velocity.V = this->velocity.x * this->velocity.x * this->movementState.x;
	this->velocity.V += this->velocity.y * this->velocity.y * this->movementState.y;
	this->velocity.V += this->velocity.z * this->velocity.z * this->movementState.z;
	
	this->velocity.V = squareRootFloat(this->velocity.V);
	
	if(this->velocity.V){
		
		this->direction.alpha.cos = FTOFIX7_9((float)abs(this->velocity.x) / this->velocity.V);
	}
	else{
		
		this->direction.alpha.cos = FTOFIX7_9(1);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// calculate character's direction angle
void Actor_calculateAnglesY(Actor this){
	
	this->velocity.V = FIX23_9_MULT(this->velocity.x, this->velocity.x) * this->movementState.x;
	this->velocity.V += FIX23_9_MULT(this->velocity.y, this->velocity.y) * this->movementState.y;
	this->velocity.V += FIX23_9_MULT(this->velocity.z, this->velocity.z) * this->movementState.z;
	
	this->velocity.V = FTOFIX23_9(squareRootFloat(FIX23_9TOF(this->velocity.V)));
	//this->direction.alpha.sin=SIN(this->direction.alpha.degree);
	
	if(this->velocity.V){
		this->direction.betha.cos = FIX23_9TOFIX7_9(FIX23_9_DIV(this->velocity.y, this->velocity.V));
	}
	else
	{
		this->direction.betha.cos = FTOFIX7_9(1);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// calculate character's direction angle
void Actor_calculateAnglesZ(Actor this){
	
	this->velocity.V = this->velocity.x * this->velocity.x * this->movementState.x;
	this->velocity.V += this->velocity.y * this->velocity.y * this->movementState.y;
	this->velocity.V += this->velocity.z * this->velocity.z * this->movementState.z;
	
	this->velocity.V = squareRootFloat(this->velocity.V);
	
	if(this->velocity.V){
		
		this->direction.tetha.cos = FTOFIX7_9((float)abs(this->velocity.z) / this->velocity.V);
	}
	else{
		
		this->direction.tetha.cos = FTOFIX7_9(1);	
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// setup velocity
void Actor_setVelocity(Actor this, int xVel, int yVel, int zVel){
	
	//set physical movement attributes
	this->velocity.x = abs(xVel);
	this->velocity.y = abs(yVel);
	this->velocity.z = abs(zVel);
	
	/*
	this->velocity.x = 0;
	this->velocity.y = 0;
	this->velocity.z = 0;
	*/
	
	Actor_calculateAnglesX(this);
	Actor_calculateAnglesY(this);
	Actor_calculateAnglesZ(this);
	//this->velocity.resultantXY=sqrtf(xVel*xVel+yVel*yVel);
	//this->velocity.resultantXZ=sqrt(xVel*xVel+zVel*zVel);
	//this->speedMultiplier=1;
	this->movementState.x = __PASSIVE;
	this->movementState.y = __PASSIVE;
	this->movementState.z = __PASSIVE;

}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set aceleration
void Actor_setAceleration(Actor this, int xAcel, int yAcel, int zAcel){
	
	//set physical movement attributes
	this->acceleration.x = xAcel;	
	this->acceleration.y = yAcel;
	this->acceleration.z = zAcel;
	
	if(!xAcel){
		this->acceleration.x = -1;		
	}
}





//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// updates the animation attributes
// graphically refresh of characters that are visible
void Actor_render(Actor this){

	// set sprite direction
	if(this->direction.x != this->previousDirection.x){
		
		// change sprite's direction
		Sprite_setDirection(this->sprite, __XAXIS, this->direction.x);
		
		// save current direction
		this->previousDirection = this->direction; 
	}

	
	//if I or the screen is moving
	if(*((int*)&this->movementState) || *((int*)_screenMovementState)){
		
		
		if(this->movementState.z | _screenMovementState->z){
		
			// calculate the scale	
			Sprite_calculateScale(this->sprite, this->position.z);
			
			// scale the sprite
			Sprite_scale(this->sprite);
			
			// set sprite's position
			Sprite_calculateParallax(this->sprite, this->position.z);
		}
		
		//update class's 2D position 
		Sprite_setPosition(this->sprite, &this->position);
		
	}	
	
	// render sprite
	AnimatedSprite_render((AnimatedSprite)this->sprite);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Actor_setInGameState(Actor this, int inGameState){
	
	this->inGameState = inGameState;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve character's movement state
MovementState Actor_getMovementState(Actor this){
	
	return this->movementState;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// execute character's logic
void Actor_update(Actor this){

	StateMachine_update(this->stateMachine);
	
	// if direction changed
	if(this->direction.x != this->previousDirection.x){
		
		// calculate gap again
		InGameEntity_setGap((InGameEntity)this);
	}
	
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve character's velocity
Velocity Actor_getVelocity(Actor this){
	
	return this->velocity;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set character's in game type
void Actor_setInGameType(Actor this, int inGameType){
	
	this->inGameType = inGameType;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// change direction
void Actor_moveOpositeDirecion(Actor this, int axis){
	
	switch(axis){
	
		case __XAXIS:
			
			this->direction.x *= -1;
			
			break;
			
		case __YAXIS:
			
			this->direction.y *= -1;
			break;
			
		case __ZAXIS:
			
			this->direction.z *= -1;
			break;
	}
	
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// update movement
void Actor_move(Actor this){
	
	if(this->movementState.x){
		Actor_keepXMovement(this);	
	}
	
	if(this->movementState.y){
		Actor_keepYMovement(this);
	}
	
	if(this->movementState.z){
		Actor_keepZMovement(this);
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// make a character to jump
void Actor_jump(Actor this, fix23_9 velocity, fix23_9 acceleration){
	
	//set the state moving over the y axis
	this->movementState.y = __ACTIVE;
	
	//initialice the dynamic changing velocity
	//set the direction over the y axis
	this->direction.y = __UP;
	
	this->velocity.y  = velocity;
	
	this->acceleration.y = acceleration;
	
	//record initial jumping time	
	this->timeY = Clock_getTime(_inGameClock);
	
	Actor_calculateAnglesY(this);
	
	this->velocity.y = this->velocity.V;

	
	//this->floorFriction=GameWorld_getFriction(_gameWorld);
	this->moveType |= __RETARMOVEY;
	
	
	if(this->moveType & __ACCELMOVEY){
		
		this->acceleration.y = (this->acceleration.y - ITOFIX23_9(__GRAVITY));
	}
	else{
		
		this->acceleration.y = FIX23_9_MULT(this->acceleration.y + ITOFIX23_9(__GRAVITY), ITOFIX23_9(-1));
	}
	
	this->maxFlyTime = this->timeY + 1000 * FIX23_9TOF(
				FIX23_9_DIV(
					FIX23_9_MULT(FIX7_9TOFIX23_9(this->direction.betha.cos), this->velocity.V), 
					ITOFIX23_9(__GRAVITY)
				)
			);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// make a character fall
void Actor_fall(Actor this){
	
	Actor_stopYMovement(this);
	
	//set the state moving over the y axis
	this->movementState.y = __ACTIVE;
	
	//set the direction over the y axis
	this->direction.y = __DOWN;
	
	//record initial falling time
	this->timeY = Clock_getTime(_inGameClock);

	this->velocity.y = ITOFIX23_9(100);
	
	Actor_calculateAnglesY(this);

	this->moveType |= __ACCELMOVEY;
	this->acceleration.y = ITOFIX23_9(__GRAVITY);
	this->timeY = Clock_getTime(_inGameClock);
	
	// if I fall, there is nothing below me
	this->objectBelow = NULL;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// start the movement over the y axis
void Actor_startYMovement(Actor this, int moveType, fix23_9 velocity, fix23_9 acceleration){
	
	//set the state moving over the y axis
	this->movementState.y=__ACTIVE;
	
	//record initial y time
	this->moveType |= moveType;
	this->timeY = Clock_getTime(_inGameClock);
	
	this->velocity.y = velocity;
	
	Actor_calculateAnglesY(this);
	
	this->velocity.y = this->velocity.V;
	//this->floorFriction=GameWorld_getFriction(_gameWorld);
	
	if(this->moveType & __ACCELMOVEY){
		
		this->acceleration.y = (this->acceleration.y - ITOFIX23_9(__GRAVITY));
	}
	else{
		
		this->acceleration.y = FIX23_9_MULT(this->acceleration.y + ITOFIX23_9(__GRAVITY), ITOFIX23_9(-1));
	}
	

	this->maxFlyTime = this->timeY + (FIX7_9TOF(this->direction.betha.cos) * this->velocity.V / __GRAVITY) * 1000;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// start the movement over the z axis
void Actor_startZMovement(Actor this, int moveType, fix23_9 velocity, fix23_9 acceleration){
	
	//set the state moving over the x axis		
	this->movementState.z = __ACTIVE;
	
	//initialice the dynamic changing velocity	
	this->velocity.z = velocity;
	
	this->acceleration.z = acceleration;
	
	//record initial x time
	this->timeZ = Clock_getTime(_inGameClock);
	
	//Actor_calculateAnglesZ(this);
	
	this->moveType |= moveType;
	
	Actor_updateAccelerationZ(this);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// start the movement over the x axis
void Actor_startXMovement(Actor this, int moveType, fix23_9 velocity, fix23_9 acceleration){
	
	//set the state moving over the x axis		
	this->movementState.x = __ACTIVE;
	
	//initialice the dynamic changing velocity	
	this->velocity.x = velocity;
	
	this->acceleration.x = acceleration;
	
	//record initial x time
	this->timeX = Clock_getTime(_inGameClock);
	
	//Actor_calculateAnglesX(this);
	
	this->moveType |= moveType;
	
	Actor_updateAccelerationX(this);
	

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// udpdate acceleration over axis
void Actor_updateAccelerationX(Actor this){

	fix7_9 friction = 0;
	
	if(this->objectBelow){
		
		friction = InGameEntity_getFrictionFactor(this->objectBelow);
	}
	
	//this->acceleration.x = this->definition->acceleration.x;
	if(this->moveType & __RETARMOVEX){
		
		this->acceleration.x += ITOFIX23_9(abs(FIX23_9TOI(this->acceleration.x - FIX23_9_MULT(this->acceleration.x, FIX7_9TOFIX23_9(friction)))));
		this->acceleration.x = FIX23_9_MULT(this->acceleration.x, ITOFIX23_9(-1));
	}
	else{
		this->acceleration.x -= ITOFIX23_9(abs(FIX23_9TOI(this->acceleration.x - FIX23_9_MULT(this->acceleration.x, FIX7_9TOFIX23_9(friction)))));
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// udpdate acceleration over axis
void Actor_updateAccelerationZ(Actor this){
	
	fix7_9 friction = 0;
	
	if(this->objectBelow){
		
		friction = InGameEntity_getFrictionFactor(this->objectBelow);
	}
	

	//this->acceleration.x = this->definition->acceleration.x;
	if(this->moveType & __RETARMOVEZ){
		
		this->acceleration.z += ITOFIX23_9(abs(FIX23_9TOI(this->acceleration.z - FIX23_9_MULT(this->acceleration.z, FIX7_9TOFIX23_9(friction)))));
		this->acceleration.z = FIX23_9_MULT(this->acceleration.z, ITOFIX23_9(-1));
	}
	else{
		this->acceleration.z -= ITOFIX23_9(abs(FIX23_9TOI(this->acceleration.z - FIX23_9_MULT(this->acceleration.z, FIX7_9TOFIX23_9(friction)))));
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// udpdate movement over axis
void Actor_keepYMovement(Actor this){

	// get the elapsed time
	fix23_9 timeElapsed = FIX23_9_DIV(ITOFIX23_9(Clock_getTime(_inGameClock) - this->timeY ) , ITOFIX23_9(1000));
	
	// the movement displacement
	fix23_9 displacement = 0;
	
	// get the angle's sin
	fix23_9 sinXP = FIX7_9TOFIX23_9(this->direction.betha.cos);
	
	// if not time has elapsed
	if(!timeElapsed){
		// do not continue processing
		return;
	}
	
	// record this update time
	this->timeY = Clock_getTime(_inGameClock);
	
	// check the movement type
	if(this->moveType & (__ACCELMOVEY | __RETARMOVEY )){
	
		// calculate the displacement in a varied movement
		displacement = FIX23_9_MULT(FIX23_9_MULT(this->velocity.y, timeElapsed), sinXP)
				+ FIX23_9_MULT(this->acceleration.y, FIX23_9_MULT(timeElapsed, timeElapsed) >> 1);
		
		// check if the character max flying time has been reached
		if((this->timeY > this->maxFlyTime) && this->direction.y == __UP){

			// stop the movement			
			Actor_stopYMovement(this);
			
			// begin to fall
			Actor_fall(this);
			// stop processing
			return;
		}
		else{
			
			// otherwise, update the velocity
			this->velocity.y += FIX23_9_MULT(this->acceleration.y, timeElapsed);			
		}
	}
	else{
		
		// movement is uniform
		displacement = FIX23_9_MULT(FIX23_9_MULT(this->velocity.V, timeElapsed), sinXP);
	}
	
	
	// update position
	this->position.y += FIX23_9_MULT(displacement, ITOFIX23_9(this->direction.y));
	
	// check if floor has been reached
	if (this->position.y >= ITOFIX23_9(__FLOOR)){
		
		MessageDispatcher_dispatchMessage(0, (Entity)this, (Entity)this, kFloorReached, NULL);
	}

}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// udpdate movement over axis
void Actor_keepXMovement(Actor this){
	
	// get the elapsed time
	fix23_9 timeElapsed = FIX23_9_DIV(ITOFIX23_9(Clock_getTime(_inGameClock) - this->timeX ) , ITOFIX23_9(1000));
	
	// the movement displacement
	fix23_9 displacement = 0;
	
	// stop processing if no time has elapsed
	if(!timeElapsed){

		return;
	}

	// determine the movement type	
	if(this->moveType & (__ACCELMOVEX | __RETARMOVEX )){
		
		// calculate displacement based in velocity, time and acceleration
		//displacement = (this->velocity.x * timeElapsed) * FIX7_9TOF(this->direction.alpha.cos) + (this->acceleration.x * timeElapsed * timeElapsed * 0.5f);
		displacement =
			FIX23_9_MULT(FIX23_9_MULT(this->velocity.x, timeElapsed), FIX7_9TOFIX23_9(this->direction.alpha.cos)) 
			+ FIX23_9_MULT(this->acceleration.x, FIX23_9_MULT(timeElapsed, timeElapsed) >> 1);
		
		// update the velocity
		this->velocity.x += FIX23_9_MULT(this->acceleration.x, timeElapsed);
		
		// if the velocity reached a very little magnitud
		if(FIX23_9TOI(this->velocity.x) < 0) {
			
			// stop movement
			Actor_stopXMovement(this);
			
			// stop processing
			return;
		}
	}
	else{
		
		// otherwise just calculate displacement in uniform movement
		displacement = FIX23_9_MULT(FIX23_9_MULT(this->velocity.x, timeElapsed), FIX7_9TOFIX23_9(this->direction.alpha.cos));
	}
	
	// update position
	this->position.x += FIX23_9_MULT(displacement, ITOFIX23_9(this->direction.x));

	// record this update's time
	this->timeX = Clock_getTime(_inGameClock);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// udpdate movement over axis
void Actor_keepZMovement(Actor this){

	// get the elapsed time
	fix23_9 timeElapsed = FIX23_9_DIV(ITOFIX23_9(Clock_getTime(_inGameClock) - this->timeZ ) , ITOFIX23_9(1000));
	
	// the movement displacement
	fix23_9 displacement = 0;
	
	// stop processing if no time has elapsed
	if(!timeElapsed){

		return;
	}
	
	// determine the movement type	
	if(this->moveType & (__ACCELMOVEZ | __RETARMOVEZ)){
		
		// calculate displacement based in velocity, time and acceleration
		displacement =
			FIX23_9_MULT(FIX23_9_MULT(this->velocity.z, timeElapsed), FIX7_9TOFIX23_9(this->direction.tetha.cos))
			+ FIX23_9_MULT(this->acceleration.z, FIX23_9_MULT(timeElapsed, timeElapsed) >> 1);
			
		
		// update the velocity
		this->velocity.z += FIX23_9_MULT(this->acceleration.z, timeElapsed);
		
		// if the velocity reached a very little magnitud
		if(FIX23_9TOI(this->velocity.z) < 0) {
			
			// stop movement
			Actor_stopZMovement(this);
			
			// stop processing
			return;
		}
	}
	else{
		
		// otherwise just calculate displacement in uniform movement
		displacement = FIX23_9_MULT(FIX23_9_MULT(this->velocity.z, timeElapsed), FIX7_9TOFIX23_9(this->direction.tetha.cos));
	}
	
	// update position
	this->position.z += FIX23_9_MULT(displacement, ITOFIX23_9(this->direction.z));

	// record this update's time
	this->timeZ = Clock_getTime(_inGameClock);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Actor_printPhysics(Actor this, int x, int y){
	
	vbjPrintText("X             Y             Z",x,y++);
	vbjPrintText("Position",x,y++);
	vbjPrintInt(this->position.x,x,y);
	vbjPrintInt(this->position.y,x+14,y);
	vbjPrintInt(this->position.z,x+14*2,y++);

	vbjPrintText("Velocity",x,y++);
	vbjPrintFloat(this->velocity.x,x,y);
	vbjPrintFloat(this->velocity.y,x+14,y);
	vbjPrintFloat(this->velocity.z,x+14*2,y++);
	vbjPrintText("Acceleration",x,y++);
	vbjPrintFloat(this->acceleration.x,x,y);
	vbjPrintFloat(this->acceleration.y,x+14,y);
	vbjPrintFloat(this->acceleration.z,x+14*2,y++);
	vbjPrintText("Direction",x,y++);
	vbjPrintInt(this->direction.x,x,y);
	vbjPrintInt(this->direction.y,x+14,y);
	vbjPrintInt(this->direction.z,x+14*2,y++);
	vbjPrintText("MovState",x,y++);
	vbjPrintInt(this->movementState.x,x,y);
	vbjPrintInt(this->movementState.y,x+14,y);
	vbjPrintInt(this->movementState.z,x+14*2,y);

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// stop movement over y axis
void Actor_stopYMovement(Actor this){
	
	this->movementState.y = __PASSIVE;
	this->timeY = 0;
	this->velocity.y = 0;
	
	//update class's 2D position 
	Sprite_setPosition(this->sprite, &this->position);

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// stop movement over z axis
void Actor_stopZMovement(Actor this){
	
	this->movementState.z = __PASSIVE;
	this->timeZ = 0;
	this->velocity.z = 0;
	
	//update class's 2D position 
	Sprite_setPosition(this->sprite, &this->position);

	// scale the sprite
	Sprite_scale(this->sprite);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// stop movement over x axis
void Actor_stopXMovement(Actor this){
	
	this->movementState.x = __PASSIVE;
	//this->lastPosition.x=this->position.x;
	this->timeX = 0;
	this->velocity.x = 0;
	
	//update class's 2D position 
	Sprite_setPosition(this->sprite, &this->position);

}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// allocate a write in graphic memory again
void Actor_resetMemoryState(Actor this, int worldLayer){		

	//Frame_resetMemoryState(this->sprite, worldLayer);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// true if inside the screen range 
int Actor_isInsideGame(Actor this){
	
	//Texture map = Sprite_getTexture(this->sprite); 
	
	return 0;//!outsideScreenRange(this->position, Texture_getCols(map), Texture_getRows(map), __CHARACTERUNLOADPAD);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// process a telegram
int Actor_handleMessage(Actor this, Telegram telegram){

	return StateMachine_handleMessage(this->stateMachine, telegram);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// stop all movement
void Actor_stopMovement(Actor this){
	
	Actor_stopXMovement(this);
	Actor_stopYMovement(this);
	Actor_stopZMovement(this);
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set movement state

void Actor_setMovementState(Actor this, MovementState movementState){
	
	this->movementState = movementState;
	

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve state machine
StateMachine Actor_getStateMachine(Actor this){
	
	return this->stateMachine;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// align character to other entity on collision
void Actor_alignX(Actor this, InGameEntity entity, int pad){

	VBVec3D position = Entity_getPosition((Entity) entity);	
	Gap gap = InGameEntity_getGap(entity);
	
	// decide to which side of the entity align myself
	if(this->position.x > position.x){
		
		// align right
		this->position.x = position.x +  
							ITOFIX23_9((Entity_getWidth((Entity)entity) >> 1) - gap.right
							+ (Entity_getWidth((Entity)this) >> 1) - this->gap.left
							+ pad);
	}
	else{
		// align left
		this->position.x = position.x - 
							ITOFIX23_9((Entity_getWidth((Entity)entity) >> 1) - gap.left 
							+ (Entity_getWidth((Entity)this) >> 1) - this->gap.right
							+ pad);

	}

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Actor_alignY(Actor this, InGameEntity entity, int pad){
	
	VBVec3D position = Entity_getPosition((Entity) entity);
	Gap gap = InGameEntity_getGap(entity);
	
	// calculate gap again
	InGameEntity_setGap((InGameEntity)this);
	
	// decide to which side of the entity align myself
	if(this->position.y > position.y){
		
		// align below
		this->position.y = position.y + 
							ITOFIX23_9((Entity_getHeight((Entity)entity) >> 1) - gap.down
							+ (Entity_getHeight((Entity)this) >> 1) - this->gap.up
							+ pad);

		// now begin to fall
		//Actor_fall(this);
		
	}
	else{
		// align over
		this->position.y = position.y - 
							ITOFIX23_9((Entity_getHeight((Entity)entity) >> 1) - gap.up 
							+ (Entity_getHeight((Entity)this) >> 1) - this->gap.down
							+ pad);
	}

		
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Actor_alignZ(Actor this, InGameEntity entity, int pad){

	VBVec3D position = Entity_getPosition((Entity) entity);
	
	// decide to which side of the entity align myself
	if(this->position.z > position.z){
		
		// align behind
		this->position.z = position.z + ITOFIX23_9(InGameEntity_getDeep(entity) + pad);
	}
	else{
		// align in front
		this->position.z = position.z - ITOFIX23_9(InGameEntity_getDeep((InGameEntity)this) + pad);

	}

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// does it moves?
int Actor_moves(Actor this){
	
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// is it moving?
int Actor_isMoving(Actor this){

	int result = 0;
	result |= (this->movementState.x)? __XAXIS: 0;
	result |= (this->movementState.y)? __YAXIS: 0;
	result |= (this->movementState.z)? __ZAXIS: 0;

	return result;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve character's scale
Scale Actor_getScale(Actor this){

	// get sprite's scale
	Scale scale = Sprite_getScale(this->sprite);
	
	// change direction
	scale.x = fabsf(scale.x) * this->direction.x;

	return scale;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// play an animation
void Actor_playAnimation(Actor this, char* animationName){
	
	AnimatedSprite_play((AnimatedSprite)this->sprite, this->actorDefinition->animationDescription, animationName);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// is play an animation
int Actor_isPlayingAnimation(Actor this, char* functionName){
	
	return AnimatedSprite_isPlayingFunction((AnimatedSprite)this->sprite, this->actorDefinition->animationDescription, functionName);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// be thrown over an axis and a direction
void Actor_beThrown(Actor this, int axis, int direction){
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// reset timers
void Actor_resetTimers(Actor this){

	//record initial x time	
	this->timeX = this->timeY = this->timeZ = Clock_getTime(_inGameClock);
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve state when unloading the entity 
int Actor_getInGameState(Actor this){

	return this->inGameState;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// the the object below me
void Actor_setObjectBelow(Actor this, InGameEntity objectBelow){
	
	this->objectBelow = objectBelow;
}