/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
 
/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												INCLUDES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#include "Rect.h"
#include "Globals.h"
#include "Optics.h"
#include "MemoryPool.h"
#include "Texture.h"


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// define the Sprite
__CLASS_DEFINITION(Rect, Shape)


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


// class's constructor
static void Rect_constructor(Rect this, InGameEntity owner, int deep);

// check if overlaps with other rect
static int Rect_overlapsRect(Rect this, Rect other, int axisMovement);


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// always call these to macros next to each other
__CLASS_NEW_DEFINITION(Rect, __PARAMETERS(InGameEntity owner, int deep))
__CLASS_NEW_END(Rect, __ARGUMENTS(this, owner, deep));

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's constructor
static void Rect_constructor(Rect this, InGameEntity owner, int deep){

	
	__CONSTRUCT_BASE(Shape,__ARGUMENTS(this, owner, deep));
	
	{
		VBVec3D position = {0, 0, 0};

		// configure the rect
		Rect_setup(this, &position);
		
		this->positionedRectangle = this->rectangle;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
void Rect_destructor(Rect this){

	// destroy the super object
	__DEALLOCATE(Shape);

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// check if two rects overlap
int Rect_overlaps(Rect this, Shape shape, int axisMovement){

	return Rect_overlapsRect(this, (Rect)shape, axisMovement);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// check if overlaps with other rect
int Rect_overlapsRect(Rect this, Rect other, int axisMovement){
	
	// must positione the rects in the owner's position
	Rectangle otherRectangle = other->positionedRectangle;

	Rectangle myRectangle = this->positionedRectangle;
	
	// get owner's position
	VBVec3D myOwnerPosition = Entity_getPosition((Entity)this->owner);
	
	VBVec3D otherOwnerPosition = Entity_getPosition((Entity)other->owner);

	// compare z positions
	if((myOwnerPosition.z  > (otherOwnerPosition.z + ITOFIX23_9(other->deep)) 
			|| 
		(myOwnerPosition.z + ITOFIX23_9(this->deep) < otherOwnerPosition.z))){
		
		return kNoCollision;
	}

	// if owner is moving
	if(axisMovement){
		
		// test for a collision
		if(!(myRectangle.x0 > otherRectangle.x1 || myRectangle.x1 < otherRectangle.x0 ||
				myRectangle.y0 > otherRectangle.y1 || myRectangle.y1 < otherRectangle.y0))
			
		{

			// this is to avoid the object being aligned in the Y axis
			// if already hit the side of the other shape
			int halfWidth = (myRectangle.x1 - myRectangle.x0) >> 2;
			
			myOwnerPosition.x = FIX23_9TOI(myOwnerPosition.x);
			myOwnerPosition.y = FIX23_9TOI(myOwnerPosition.y);

			// now guess over which axis the collision is
			// giving preference to collisions over the x axis
			if((__XAXIS & axisMovement) &&
					(myOwnerPosition.x + halfWidth <= otherRectangle.x0 
							|| 
						myOwnerPosition.x - halfWidth >= otherRectangle.x1)){

				return kCollisionX;
			}
			
			if(__ZAXIS & axisMovement){
				
				return kCollisionZ;
			}
			
			if((__YAXIS & axisMovement) &&
			    (myOwnerPosition.y < otherRectangle.y0 
			    || 
			    myOwnerPosition.y > otherRectangle.y1)){
				
				return kCollisionY;
			}
					
			return kCollisionXY;
		}
	}
	
	// if not moving or a collision was not detected
	// check if the shape is below
	if(!(myRectangle.x0 > otherRectangle.x1 || myRectangle.x1 < otherRectangle.x0)){

		myRectangle.y1 += 1;
		
		if(myRectangle.y1 >= otherRectangle.y0 && myRectangle.y0 < otherRectangle.y0){
			
			return kShapeBelow;	
		}
	}
	
	return kNoCollision;

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Rect_setup(Rect this, const VBVec3D* const position){
	
	// rect's center if placed on P(0, 0)
	this->rectangle.x1 = Entity_getWidth((Entity)this->owner) >> 1;
	this->rectangle.y1 = Entity_getHeight((Entity)this->owner) >> 1;
	
	this->rectangle.x0 = -this->rectangle.x1;
	this->rectangle.y0 = -this->rectangle.y1;
	
	// if owner does not move
	if(!this->moves){
		
		// position the shape to avoid in real time calculation
		VBVec3D ownerPosition = Entity_getPosition((Entity)this->owner);
		Gap ownerGap = InGameEntity_getGap(this->owner);

		// calculate gap on each side of the rectangle
		this->rectangle.x0 += FIX23_9TOI(ownerPosition.x) + ownerGap.left;
		this->rectangle.x1 += FIX23_9TOI(ownerPosition.x) - ownerGap.right;
		this->rectangle.y0 += FIX23_9TOI(ownerPosition.y) + ownerGap.up;
		this->rectangle.y1 += FIX23_9TOI(ownerPosition.y) - ownerGap.down;
	}
	
	this->positionedRectangle = this->rectangle;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// prepare the shape to be checked
void Rect_prepare(Rect this){
		
	Gap gap = InGameEntity_getGap(this->owner);

	// get owner's position
	VBVec3D myOwnerPosition = Entity_getPosition((Entity)this->owner);

	// calculate positioned rectangle
	myOwnerPosition.x = FIX23_9TOI(myOwnerPosition.x);
	myOwnerPosition.y = FIX23_9TOI(myOwnerPosition.y);
	
	this->positionedRectangle = this->rectangle;
	
	this->positionedRectangle.x0 += myOwnerPosition.x + gap.left;
	this->positionedRectangle.x1 += myOwnerPosition.x - gap.right;
	this->positionedRectangle.y0 += myOwnerPosition.y + gap.up;
	this->positionedRectangle.y1 += myOwnerPosition.y - gap.down;

	// not checked yet
	this->checked = false;
}


