/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												INCLUDES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#include "TextureManager.h"
#include "Error.h"
#include "Constants.h"
#include "MemoryPool.h"


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											 CLASS'S MACROS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

//bgmaps to use (leave 2 bgmaps to allocate param table)
#define NUM_BGMAPS 12

//number of bgmap definitions in each bgmap segment 
#define NUM_MAPS_PER_SEG 16


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#define TextureManager_ATTRIBUTES											\
																		\
	/* it is derivated from */											\
	Object_ATTRIBUTES													\
																		\
	/* number of chars ocuppied */										\
	int numberOfChars[NUM_BGMAPS];										\
																		\
	/* current x offset to set the next bgmap */						\
	int xOffset[NUM_BGMAPS][NUM_MAPS_PER_SEG];							\
																		\
	/* current y offset to set the next bgmap */						\
	int yOffset[NUM_BGMAPS][NUM_MAPS_PER_SEG];							\
																		\
	/* 12 segments, 28 maps, 2 indexes (x,y) and bgmap segment */ 		\
	int offset[NUM_BGMAPS * NUM_MAPS_PER_SEG][3];						\
																		\
	/* bit mask to know where is the next free bgmap */					\
	/* int bgmapUsage; */												\
																		\
	/* next free bgmap used for text printing */						\
	int freeBgmap;														\
																		\
	/* allocated bgmap's definition*/									\
	BYTE *definition[NUM_BGMAPS * NUM_MAPS_PER_SEG];					\
																		\
	/* usage for each bgmap's definition */								\
	BYTE  definitionUsage[NUM_BGMAPS * NUM_MAPS_PER_SEG];				

	

// define the AnimatedSprite
__CLASS_DEFINITION(TextureManager, Object)

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// class's constructor
static void TextureManager_constructor(TextureManager this);

// record the bgmap's definition within the TextureManager class
static void TextureManager_recordEntry(TextureManager this, BYTE* bgmapDef, int segment, int xOffset, int yOffset);

// search bgmap's definition in the TextureManager class
static int TextureManager_searchMapDefinition(TextureManager this, Texture texture);

// search bgmap's shared definition in the TextureManager class
static int TextureManager_searchSharedMapDefinition(TextureManager this, Texture texture);

// determine if a bgmap exclusive use definition is free
static int TextureManager_searchFreeMapDefinition(TextureManager this, Texture texture);

// allocate texture in bgmpa graphic memory
static int TextureManager_allocate(TextureManager this, Texture texture);



/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */




//////////////////////////////////////////////////////////////////////////////////////////////////////////////

__SINGLETON(TextureManager);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's constructor
static void TextureManager_constructor(TextureManager this){
	
	__CONSTRUCT_BASE(Object,__ARGUMENTS(this));
	{
		int i = 0;
		int j = 0;
	
		// clear each bgmap segment usage
		for(; i < NUM_BGMAPS; i++){
			
			this->numberOfChars[i] = 0;
			
			// clear the offsets
			for(j=0;j<NUM_MAPS_PER_SEG;j++){
				
				this->xOffset[i][j] = 0;
				this->yOffset[i][j] = 0;
				
				this->xOffset[i][0] = 1;
				this->yOffset[i][0] = 1;
			}
		}
		
		for(i = 0; i < NUM_BGMAPS * NUM_MAPS_PER_SEG; i++){
			
			this->offset[i][0] = 0;
			this->offset[i][1] = 0;
			this->offset[i][2] = 0;
			this->definition[i] = NULL;
			this->definitionUsage[i] = 0;
		}
		this->freeBgmap = 0;
	}
	//this->bgmapUsage = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
void TextureManager_destructor(TextureManager this){

	// allow a new construct
	__SINGLETON_DESTRUCT;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// deallocate texture from bgmap graphic memory
void TextureManager_free(TextureManager this, Texture texture){
	
	//get texture's definition index within class's bgmap definition array
	int i = TextureManager_searchMapDefinition(this, texture);

	//an undefinded definition is being deallocated
	ASSERT(i >= 0, BGMEM_NOALLOC_ERR);
		
	//decrease definition usage
	this->definitionUsage[i]-=1;
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// allocate texture in bgmap graphic memory
static int TextureManager_allocate(TextureManager this, Texture texture){
	
	int i = 0;
	int j = 0;	
	int aux = 0;
	//int bgmapMask = 0x0001;
	
	int rows = Texture_getRows(texture) + 1;
	int cols = Texture_getTotalCols(texture) + 1;
	
	int area = rows * cols;
	
	//if texture already defined, don't allocate
	CACHE_ENABLE;	
	if(Texture_getNumberOfChars(texture)){
		
		for(i = 0; i < NUM_BGMAPS; i++){
			//if there is space in the segment memory
			// there are 4096 chars in each bgmap segment
			if((4096 - this->numberOfChars[i]) >= area ){
				
				//set the texture's bgmap segment number
				//Texture_setBgmapSegment(texture,i);
				
				//check if there is space within the segment
				// we check the next so don't go to the last element
				for(j = 0; j < NUM_MAPS_PER_SEG - 1; j++){

					//determine the y offset inside the bgmap segment
					if(!this->yOffset[i][j + 1]){
						
						aux = 64;
					}
					else{
						
						aux = this->yOffset[i][j + 1];

					}
					
					//determine if there is still mem space (columns) in the current y offset
					if(rows < aux - this->yOffset[i][j] || (!this->yOffset[i][j+1])){
						
						if(rows < 64 - this->yOffset[i][j]){
						
							if(cols < 64 - this->xOffset[i][j]){
								
								//set texture's yoffset
								Texture_setYOffset(texture, this->yOffset[i][j]);
								
								//set texture's xoffset
								Texture_setXOffset(texture, this->xOffset[i][j]);
								
								//registry bgmap definition
								TextureManager_recordEntry(this, Texture_getBgmapDef(texture), i, this->xOffset[i][j], this->yOffset[i][j]);
								
								//increment the x offset
								this->xOffset[i][j] += cols;
								
								//if the number of rows of the bgmap definition is greater than the
								//next y offset defined, increase the next y offset
								if(this->yOffset[i][j+1] - this->yOffset[i][j] < rows){
									
									this->yOffset[i][j+1] = this->yOffset[i][j] + rows ;
								}
								
								//update the number of chars defined inside the bgmap segment
								this->numberOfChars[i] += area;
								
								//set bgmap to print debug information
								Texture_setBgmapSegment(texture, i);
								
								if(i + 1 > this->freeBgmap){
									
									this->freeBgmap = i + 1;
								}

								
								CACHE_DISABLE;
								
								//if there is a free bgmap segment
								return true;
							}
						}
						else
							if(rows > 64 - this->yOffset[i][j]){
								j = NUM_MAPS_PER_SEG - 1;
								break;}

					}	
					else{
						if(rows > 64 - this->yOffset[i][j]){
							j = NUM_MAPS_PER_SEG - 1;
							break;
						}
					}
				}
			}
		}

		//through exception if there is no enough space to allocate the bgmap definition
		ASSERT(false, BGMEM_MEM_ERR);
		
	}
	CACHE_DISABLE;

	//through exception if texture has 0 chars
	ASSERT(false, BGMEM_0CH_ERR);
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// for no animated textures
int TextureManager_allocateShared(TextureManager this, Texture texture){
	
	int i = TextureManager_searchSharedMapDefinition(this, texture);
	
	//verify that the char's definition is not already defined		
	if(i < 0){
		//if not, then allocate
		TextureManager_allocate(this, texture);
		
		return true;		
	}
	else{
		// set texture's segment and offset
		Texture_setXOffset(texture, this->offset[i][0]);
		Texture_setYOffset(texture, this->offset[i][1]);
		Texture_setBgmapSegment(texture, this->offset[i][2]);
		
		//set texture's states as "in use"
		this->definitionUsage[i]++;
		
		//if already defined, but is the first time to being use
		if(this->definitionUsage[i] == 1){			
		
			//make it to be writen to memory again because of char rellocation
			return true;
		}
		
		return false;
	}
}
	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// for animated texture characters
int TextureManager_allocateDedicated(TextureManager this, Texture texture){
	
	int i = TextureManager_searchFreeMapDefinition(this, texture);
	
	//verify that the char's definition is not already defined		
	if(i < 0){
		//if not, then allocate
		TextureManager_allocate(this, texture);
	}
	else{		
		// set texture's segment and offset
		Texture_setXOffset(texture, this->offset[i][0]);
		Texture_setYOffset(texture, this->offset[i][1]);
		Texture_setBgmapSegment(texture, this->offset[i][2]);
		
		//set texture's states as "in use"
		this->definitionUsage[i] = 1;
	}
	return true;	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve free bgmap segment number
int TextureManager_getFreeBgmap(TextureManager this){
	
	return this->freeBgmap;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// allocate bgmap text boxes
// this bgmap segment is handled as one only bgmap defined inside so, only
// TextureManager.xOffset[textbgmap][0] is used
void TextureManager_allocateText(TextureManager this, Texture texture){
	
	int xDisplacement = 0;
	int yDisplacement = 0;
	
	int length = Texture_getCols(texture);
	
	//if there is space in the first row
	//calculate y displacement
	//offset/64->chars per row inside a bgmap
	yDisplacement = (this->xOffset[this->freeBgmap][0] + length) >> 6;
	
	//move to the next row
	if(this->xOffset[this->freeBgmap][0] < 64 * yDisplacement){
		this->xOffset[this->freeBgmap][0] = 64 * yDisplacement;
	}
	
	//offset%/64->chars per row inside a bgmap
	xDisplacement = (this->xOffset[this->freeBgmap][0]) % 64;
	
	//let free the upper chars to debug and quick text printing
	Texture_setYOffset(texture, 28 + yDisplacement);
	
	//calculate x displacement
	Texture_setXOffset(texture, xDisplacement);
	
	//set next ofsset entry to modify within the free bgmap segment
	this->xOffset[this->freeBgmap][0] += length;

	Texture_setBgmapSegment(texture, this->freeBgmap);

	//if there are no more rows in the segment... thrown and exception
	ASSERT(this->xOffset[this->freeBgmap][0] < 4096, BGMEM_TX_ERR);
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// record the bgmap's definition within the TextureManager class
static void TextureManager_recordEntry(TextureManager this, BYTE* bgmapDef, int segment, int xOffset, int yOffset){
	
	int i = 0;
	
	for(;i < NUM_BGMAPS * NUM_MAPS_PER_SEG && this->definition[i]; i++);
		
	ASSERT(i < NUM_BGMAPS * NUM_MAPS_PER_SEG, TextureManager: not defition space left);
			
	this->definition[i] = bgmapDef;
	this->offset[i][0] = xOffset;
	this->offset[i][1] = yOffset;
	this->offset[i][2] = segment;
	//texture's use state
	this->definitionUsage[i] = 1;
}			

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// search bgmap's definition in the TextureManager class
static int TextureManager_searchMapDefinition(TextureManager this, Texture texture){
	int i = 0;
	void* bgmapDef = Texture_getBgmapDef(texture);
	
	for(; i < NUM_BGMAPS * NUM_MAPS_PER_SEG; i++){
		
		//if definition is the same and not in use
		if(this->definition[i] == bgmapDef){
			
			if(Texture_getXOffset(texture) == this->offset[i][0] && Texture_getYOffset(texture) == this->offset[i][1]){
				
				return i;
			}
		}
	}
	return -1;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// search bgmap's shared definition in the TextureManager class
static int TextureManager_searchSharedMapDefinition(TextureManager this, Texture texture){
	int i = 0;
	void* bgmapDef = Texture_getBgmapDef(texture);
	
	for(; i < NUM_BGMAPS * NUM_MAPS_PER_SEG; i++){
		
		//if definition is the same and not in use
		if(this->definition[i] == bgmapDef){
			
			return i;
		}
	}
	return -1;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// determine if a bgmap exclusive use definition is free
static int TextureManager_searchFreeMapDefinition(TextureManager this, Texture texture){
	
	int i = 0;
	
	void* bgmapDef = Texture_getBgmapDef(texture);	
	
	for(;i < NUM_BGMAPS * NUM_MAPS_PER_SEG; i++){
		
		//if definition is the same and not in use
		if(this->definition[i] == bgmapDef && !this->definitionUsage[i]){
			
			return i;
		}
	}
	return -1;
}


