/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
 
/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												INCLUDES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#include "Entity.h"
#include "Globals.h"
#include "Constants.h"
#include "Optics.h"
#include "MemoryPool.h"
#include "../config.h"


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// define the Sprite
__CLASS_DEFINITION(Entity, Object)


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */




/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's constructor
void Entity_constructor(Entity this, EntityDefinition* entityDefinition, int ID){

	// this is an abstract class so must initialize the vtable here
	// since this class does not have an allocator
	__SET_CLASS(Entity);	

	// construct base object
	__CONSTRUCT_BASE(Object,__ARGUMENTS(this));

	/* the sprite must be initializated in the derivated class */
	this->sprite = NULL;
	
	// set ID
	this->ID = ID;
	
	// set position
	this->position.x = 0;
	this->position.y = 0;
	this->position.z = 0;
		
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
void Entity_destructor(Entity this){


	ASSERT(this->sprite);
		
	// delete the sprite
	__DELETE(this->sprite);

	// destroy the super object
	__DEALLOCATE(Object);

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// process extra info in intialization
void Entity_setExtraInfo(Entity this, void* extraInfo){
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// update
void Entity_update(Entity this){
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// write Entity to graphic memory
void Entity_write(Entity this){
	
	ASSERT(this->sprite);
	Sprite_write(this->sprite);
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// allocate a write in graphic memory again
void Entity_resetMemoryState(Entity this, int worldLayer){
	
	// make sure is is allocated and written
	//Map_resetMemoryState(this->map);
	
	//Map_setWorldLayer(this->map, worldLayer);

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// render class
void Entity_render(Entity this){

	ASSERT(this->sprite);
	
	//if screen is moving
	if(*((int*)_screenMovementState)){
		
		if(_screenMovementState->z){

			// calculate the scale	
			Sprite_calculateScale(this->sprite, this->position.z);
			
			// scale the sprite
			Sprite_scale(this->sprite);			

			// calculate sprite's parallax
			Sprite_calculateParallax(this->sprite, this->position.z);
		}
		
		//update class's 2D position 
		Sprite_setPosition(this->sprite, &this->position);
	}	
	
	//render map	
	__VIRTUAL_CALL(void, Sprite, render, this->sprite);

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set class's position
void Entity_setPosition(Entity this, VBVec3D position){
	
	this->position = position;
	
	//update class's 2D position 
	Sprite_setPosition(this->sprite, &this->position);
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve class's in game index
int Entity_getID(Entity this){

	return this->ID;
}

// retrieve class's scale
Scale Entity_getScale(Entity this){
	
	ASSERT(this->sprite);
	
	return Sprite_getScale(this->sprite);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve position
inline VBVec3D Entity_getPosition(Entity this){

	return this->position;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve sprite
Sprite Entity_getSprite(Entity this){

	ASSERT(this->sprite);
	
	return this->sprite;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// process a telegram
int Entity_handleMessage(Entity this, Telegram telegram){
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get width
int Entity_getWidth(Entity this){

	Texture texture = Sprite_getTexture(this->sprite);
	
	ASSERT(this->sprite);
	
	// must calculate based on the scale because not affine object must be enlarged
	return vbjCalculateRealSize(Texture_getCols(texture) << 3, Sprite_getMode(this->sprite), abs(Sprite_getScale(this->sprite).x));
}



//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get height
int Entity_getHeight(Entity this){

	Texture texture = Sprite_getTexture(this->sprite);
	
	ASSERT(this->sprite);
	
	// must calculate based on the scale because not affine object must be enlarged
	return vbjCalculateRealSize(Texture_getRows(texture) << 3, Sprite_getMode(this->sprite), abs(Sprite_getScale(this->sprite).y));
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// whether it is visible
int Entity_isVisible(Entity this, int pad){
	
	ASSERT(this->sprite);
	
	return vbjIsVisible(this->position,
			Entity_getWidth(this),
			Entity_getHeight(this),
			Sprite_getDrawSpec(this->sprite).position.parallax,
			pad);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// determine if the entity is outside the game
int Entity_isOutsideGame(Entity this){
	
	ASSERT(this->sprite);
	
	return vbjOutsideGame(this->position, Entity_getWidth(this), Entity_getHeight(this));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve state when unloading the entity 
int Entity_getInGameState(Entity this){

	return kUnloaded;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// update sprite's world layer 
void Entity_updateLayer(Entity this, int worldLayerIncrement){

	ASSERT(this->sprite);
	
	Sprite_setWorldLayer(this->sprite, Sprite_getWorldLayer(this->sprite) + worldLayerIncrement);
		
	Sprite_render(this->sprite);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set sprite's world layer 
void Entity_setLayer(Entity this, int worldLayer){

	ASSERT(this->sprite);
	
	Sprite_setWorldLayer(this->sprite, worldLayer);
		
	//Sprite_render(this->sprite);
}