/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												INCLUDES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#include "Sprite.h"
#include "Globals.h"
#include "MessageDispatcher.h"
#include "MemoryPool.h"
#include "Printing.h"
#include "Optics.h"
#include "../config.h"



/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											 CLASS'S MACROS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// define the Sprite
__CLASS_DEFINITION(Sprite, Object)


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */





/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */
		


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// always call these to macros next to each other
__CLASS_NEW_DEFINITION(Sprite, __PARAMETERS(SpriteDefinition* spriteDefinition))
__CLASS_NEW_END(Sprite, __ARGUMENTS(this, spriteDefinition));

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's conctructor
void Sprite_constructor(Sprite this, SpriteDefinition* spriteDefinition){

	__CONSTRUCT_BASE(Object,__ARGUMENTS(this));
	
	{
		
		// create the texture			 
		this->texture = __NEW(Texture, __ARGUMENTS(spriteDefinition->textureDefinition));

		// clear position
		this->drawSpec.position.x = 0;
		this->drawSpec.position.y = 0;
		this->drawSpec.position.parallax = 0;
		
		this->drawSpec.scale.x = ITOFIX7_9(1);
		this->drawSpec.scale.y = ITOFIX7_9(1);

		this->param = 0;
			
		//this->head = spriteDefinition->display | WRLD_BGMAP;	
		//set world layer's head acording to map's render mode
		switch(spriteDefinition->bgmapMode){
		
			case WRLD_BGMAP:
				
				//set map head
				this->head = spriteDefinition->display | WRLD_BGMAP;	
				
				break;
				
			case WRLD_AFFINE:
		
				//set map head			
				this->head = spriteDefinition->display | WRLD_AFFINE | WRLD_OVR;
				
				//allocate param table space			
				ParamTableManager_allocate(ParamTableManager_getInstance(), this);
				
				break;
				
			case WRLD_HBIAS:
				
				//set map head
				this->head = spriteDefinition->display | WRLD_HBIAS | WRLD_OVR;
				
				break;
		}
	

		// set the default layer
		this->worldLayer = 0;
		
		// set the render flag
		this->renderFlag = __UPDATEHEAD;
			
		// write the texture
		//Texture_write(this);	
		this->updateParamTable = true;


	}		
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
void Sprite_destructor(Sprite this){


	// delete the sprite
	__DELETE(this->texture);

	
	//if affine or bgmap
	if(WRLD_AFFINE & this->head){

		//free param table space
		ParamTableManager_free(ParamTableManager_getInstance(), this);
	}	
	
	// destroy the super object
	__DEALLOCATE(Object);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// write background to graphic memory
void Sprite_write(Sprite this){

	// write the texture
	Texture_write(this->texture);	

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
Scale Sprite_getScale(Sprite this){
	
	//  return the scale
	return this->drawSpec.scale;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set the direction
void Sprite_setDirection(Sprite this, int axis, int direction){
	
	switch(axis){
	
		case __XAXIS:
			
			this->drawSpec.scale.x = FIX7_9_MULT(abs(this->drawSpec.scale.x), ITOFIX7_9(direction));
			
			break;
			
		case __YAXIS:
					
			this->drawSpec.scale.y = FIX7_9_MULT(abs(this->drawSpec.scale.y), ITOFIX7_9(direction));
			break;
	}
	
	// scale the texture in the next render cycle
	Sprite_invalidateParamTable(this);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// calculate zoom scaling factor
void Sprite_calculateScale(Sprite this, fix23_9 z){
	
	fix7_9 ratio = FIX23_9TOFIX7_9(ITOFIX23_9(1) - 
			       FIX23_9_DIV(z , _optical->maximunViewDistance));

	// TODO: remove the * and the branch
	this->drawSpec.scale.x = ratio * (this->drawSpec.scale.x < 0? -1: 1);
	this->drawSpec.scale.y = ratio;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set sprite's position
void Sprite_setPosition(Sprite this, const VBVec3D* const position){

	// normalize the position to screen coordinates
	VBVec3D position3D = vbjNormalizePosition(position);

	// project position to 2D space
	vbjProjectTo2D(&this->drawSpec.position, &position3D);
	
	// make the texture being rendered in the next render loop
	this->renderFlag |=  __UPDATEG;
	
	// if texture is in bgmap mode
	if(WRLD_AFFINE != Sprite_getMode(this)){

		// now move the drawspec in order to render the texture in the center
		this->drawSpec.position.x -= (Texture_getCols(this->texture) << 2);
		this->drawSpec.position.y -= (Texture_getRows(this->texture) << 2);

	}
	else{

		// now move the drawspec in order to render the texture in the center
		this->drawSpec.position.x -= FIX23_9_ROUNDTOI(
				FIX23_9_MULT(ITOFIX23_9(Texture_getCols(this->texture) << 2), FIX7_9TOFIX23_9(abs(this->drawSpec.scale.x)))
				);
		
		this->drawSpec.position.y -= FIX23_9_ROUNDTOI(
				FIX23_9_MULT(ITOFIX23_9(Texture_getRows(this->texture) << 2), FIX7_9TOFIX23_9((this->drawSpec.scale.y)))
				);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// calculate the parallax
void Sprite_calculateParallax(Sprite this, fix23_9 z){
	
	this->drawSpec.position.parallax = vbjCalculateParallax(this->drawSpec.position.x, z);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve the texture
Texture Sprite_getTexture(Sprite this){
	
	return this->texture;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve drawspec
DrawSpec Sprite_getDrawSpec(Sprite this){
	
	return this->drawSpec;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set to true to allow render
void Sprite_setRenderFlag(Sprite this, int renderFlag){
	
	// do not override it the whole world entry must be updated in the
	// next render
	if(__UPDATEHEAD != this->renderFlag || !renderFlag) {
		
		this->renderFlag = renderFlag;
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// render a world layer with the map's information
void Sprite_render(Sprite this){

	//if render flag is set
	if(this->renderFlag){
		
		DrawSpec drawSpec = this->drawSpec;

		if(__UPDATEHEAD == this->renderFlag){
			
			//create an independant of software variable to point XPSTTS register
			unsigned int volatile *xpstts =	(unsigned int *)&VIP_REGS[XPSTTS];
			
			//wait for screen to idle	
			while (*xpstts & XPBSYR);
			
			// write the head
			WORLD_HEAD(this->worldLayer, this->head | Texture_getBgmapSegment(this->texture));

			WORLD_MSET(this->worldLayer, (Texture_getXOffset(this->texture) << 3), 0, Texture_getYOffset(this->texture) << 3);
		}
		
		//set the world screen position
		if(this->renderFlag & __UPDATEG ){

			WORLD_GSET(this->worldLayer, drawSpec.position.x, drawSpec.position.parallax, drawSpec.position.y);
		}
		
		//set the world size according to the zoom
		if(WRLD_AFFINE & this->head){

			// check if must update the param table
			if(this->updateParamTable){

				// now scale the texture
				Sprite_scale(this);

				// don't update on next render cycle
				this->updateParamTable = false;
			}
			
			if(this->renderFlag & __UPDATESIZE){

				WORLD_SIZE(this->worldLayer, 
						FIX23_9_ROUNDTOI(FIX7_9TOFIX23_9(FIX7_9_MULT(ITOFIX7_9(Texture_getCols(this->texture)<< 3), abs(drawSpec.scale.x)))) - 1,						
						FIX23_9_ROUNDTOI(FIX7_9TOFIX23_9(FIX7_9_MULT(ITOFIX7_9(Texture_getRows(this->texture)<< 3), abs(drawSpec.scale.y)))));
				
				
			}
			
			//set the world paralax
			if(this->renderFlag & __UPDATEPARAM){
				
				WORLD_PARAM(this->worldLayer, PARAM(this->param));				
			}

		}
		else{
			
			if(this->renderFlag & __UPDATESIZE){
				
				WORLD_SIZE(this->worldLayer, (Texture_getCols(this->texture) << 3), (Texture_getRows(this->texture) << 3));
			}
			
			if(this->renderFlag & __UPDATEM ){
				
				//set the world cuting bgmap memory point
				WORLD_MSET(this->worldLayer, (Texture_getXOffset(this->texture) << 3), 0, Texture_getYOffset(this->texture) << 3);
			}
		}

		// make sure to not render again
		this->renderFlag = false;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// render a world layer with the map's information
void Sprite_renderAll(Sprite this){

	DrawSpec drawSpec = this->drawSpec;
	
	//create an independant of software variable to point XPSTTS register
	unsigned int volatile *xpstts =	(unsigned int *)&VIP_REGS[XPSTTS];

	//wait for screen to idle	
	while (*xpstts & XPBSYR);
	
	// write the head
	WORLD_HEAD(this->worldLayer, this->head | Texture_getBgmapSegment(this->texture));

	WORLD_GSET(this->worldLayer, drawSpec.position.x, drawSpec.position.parallax, drawSpec.position.y);
	
	//set the world size according to the zoom
	if(WRLD_AFFINE & this->head){

		// check if must update the param table
		if(this->updateParamTable){

			// now scale the texture
			Sprite_scale(this);

			// don't update on next render cycle
			this->updateParamTable = false;
		}
		

		WORLD_SIZE(this->worldLayer, 
				FIX23_9_ROUNDTOI(FIX7_9TOFIX23_9(FIX7_9_MULT(ITOFIX7_9(Texture_getCols(this->texture)<< 3), abs(drawSpec.scale.x)))) - 1,						
				FIX23_9_ROUNDTOI(FIX7_9TOFIX23_9(FIX7_9_MULT(ITOFIX7_9(Texture_getRows(this->texture)<< 3), abs(drawSpec.scale.y)))));
			
			
		WORLD_PARAM(this->worldLayer, PARAM(this->param));				
	}
	else{
		
		WORLD_SIZE(this->worldLayer, (Texture_getCols(this->texture) << 3), (Texture_getRows(this->texture) << 3));

		//set the world cuting bgmap memory point
		WORLD_MSET(this->worldLayer, (Texture_getXOffset(this->texture) << 3), 0, Texture_getYOffset(this->texture) << 3);
		
	}

	this->renderFlag = __UPDATEHEAD;

}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get map's param table address
u32 Sprite_getParam(Sprite this){
	
	return this->param;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set map's param table address
void Sprite_setParam(Sprite this, u32 param){

	this->param = param;
	
	// set flag to rewrite texture's param table
	Sprite_invalidateParamTable(this);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set map's world layer
void Sprite_setWorldLayer(Sprite this, int worldLayer){

	if(this->worldLayer != worldLayer){
	
		this->worldLayer = worldLayer;
	
		this->renderFlag = __UPDATEHEAD;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//get map's world layer
int Sprite_getWorldLayer(Sprite this){
		
	return this->worldLayer;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get sprite's render head
int Sprite_getHead(Sprite this){
	
	return this->head;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get map's render mode
int Sprite_getMode(Sprite this){
	
	return this->head & 0x3000;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve param table flag
int Sprite_updateParamTable(Sprite this){
	
	return this->updateParamTable;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// force refresh param table in the next render
void Sprite_invalidateParamTable(Sprite this){
	
	this->updateParamTable = true;
	this->renderFlag |= __UPDATESIZE | __UPDATEPARAM | __UPDATEHEAD;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// this reallocate a write the bgmap definition in graphical memory
void Sprite_resetMemoryState(Sprite this){
	
	//if affine or hbias mode, allocate inside paramtable		
	if(WRLD_AFFINE == Sprite_getMode(this)){
		
		ParamTableManager_allocate(ParamTableManager_getInstance(), this);
		
	}
	
	//allow to render
	this->renderFlag = __UPDATEHEAD;
	
	// write it in graphical memory
	Texture_resetMemoryState(this->texture);
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set drawspec
void Sprite_setDrawSpec(Sprite this, const DrawSpec* const drawSpec){

	this->drawSpec = *drawSpec;
	
}
/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 										MAP FXs
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

/*
 * Affine FX
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Sprite_noAFX(Sprite this, int direction){
	
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Sprite_scale(Sprite this){
	
	//put the map into memory calculating the number of char for each reference
	if(this->param){
		
		int cols = Texture_getCols(this->texture) << 2;
		int rows = Texture_getRows(this->texture) << 2;

		affineScale(this->param, this->drawSpec.scale.x, this->drawSpec.scale.y, 
				   (Texture_getXOffset(this->texture) << 3) + cols,
				   (Texture_getYOffset(this->texture) << 3) + rows,
				   cols, rows);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Sprite_rotate(Sprite this, int angle){
	// TODO
	//put the map into memory calculating the number of char for each reference
	//put the map into memory calculating the number of char for each reference
	if(this->param){
		
		int cols = Texture_getCols(this->texture) << 2;
		int rows = Texture_getRows(this->texture) << 2;

		affineRotateZ(this->param, this->drawSpec.scale.x, this->drawSpec.scale.y, 
				   (Texture_getXOffset(this->texture) << 3) + cols,
				   (Texture_getYOffset(this->texture) << 3) + rows,
				   cols, rows, angle);
	
		
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// TODO
	/*
	static int alpha=0;
	if(this->updateParamTable==true){
		affineRotateY(this->param,alpha,this->scale.x,this->scale.y, 
		(this->xOffset<<3)+(this->cols<<2), (this->yOffset<<3)+(this->rows<<2), 
		(this->cols<<2),(this->rows<<2));
		if(alpha++ >125){
			alpha=125;
					
		}
		// put down the flag
		Sprite_setUpdateParamTableFlag(this, false);
		
	}
	*/
	//delay(5);

/*
 * H-Bias FX
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Sprite_squezeXHFX(Sprite this){
	// TODO
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Sprite_fireHFX(Sprite this){
	// TODO
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Sprite_waveHFX(Sprite this){
	// TODO
}



