/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												INCLUDES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#include "Texture.h"
#include "Globals.h"
#include "Constants.h"
#include "Optics.h"
#include "Printing.h"


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											 CLASS'S MACROS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


#define PALLET_MASK			0x0600
#define WORLD_LAYER_MASK	0x01F0
#define SEGMENT_MASK		0x000F


#define PALLET_MASK_DISP		0x09 /* 6 */
#define WORLD_LAYER_MASK_DISP	0x04 /* 1 */
#define SEGMENT_MASK_DISP		0x00 /* 0 */


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */




__CLASS_DEFINITION(Texture, Object);

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// class's constructor
static void Texture_constructor(Texture this, TextureDefinition* textureDefinition);

// write an animated map
static void Texture_writeAnimated(Texture this);

// write an inanimated map
static void Texture_writeNoAnimated(Texture this);

// write an animated and shared map
static void Texture_writeAnimatedShared(Texture this);


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// always call these to macros next to each other
__CLASS_NEW_DEFINITION(Texture, __PARAMETERS(TextureDefinition* textureDefinition))
__CLASS_NEW_END(Texture, __ARGUMENTS(this, textureDefinition));


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's constructor
static void Texture_constructor(Texture this, TextureDefinition* textureDefinition){

	// construct base object
	__CONSTRUCT_BASE(Object,__ARGUMENTS(this));

	// if the char definition is NULL, it must be a text
	this->charGroup = __NEW(CharGroup, __ARGUMENTS(&textureDefinition->charGroupDefinition));
	
	// save the bgmap definition's address
	this->bgmapDef = textureDefinition->bgmapDefinition;	
	
	// set texture's size
	this->cols = textureDefinition->cols;
	this->rows = textureDefinition->rows;
	
	// set texture's displacement inside bgmap memory
	this->xOffset=(int)0xFF;
	this->yOffset=(int)0xFF;
	
	// set the pallet
	this->pallet = textureDefinition->pallet;

	// set the default bgmap segment
	this->bgmapSegment = 0;		
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
void Texture_destructor(Texture this){
	
	//destroy the chargroup
	__DELETE(this->charGroup);
	
	//free bgmap mem
	TextureManager_free(TextureManager_getInstance(), this);
	
	// destroy the super object
	__DEALLOCATE(Object);
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// write an animated map
static void Texture_writeAnimated(Texture this){

	int bgmapSegment = Texture_getBgmapSegment(this);
	int pallet = Texture_getPallet(this) << 14;
	
	int charLocation = (CharGroup_getCharSet(this->charGroup) << 9) + CharGroup_getOffset(this->charGroup);
	int i = this->rows;

	//put the map into memory calculating the number of char for each reference
	for(; i--;){
		
		// write each row to graphic memory
		addMem ((void*)BGMap(bgmapSegment) + ((this->xOffset + (this->yOffset << 6 ) + (i << 6)) << 1), 
				(const u16*)(this->bgmapDef + ( i * (this->cols) << 1)), 
				this->cols,
				(pallet) | (charLocation));
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// write an inanimated map
static void Texture_writeNoAnimated(Texture this){

	int bgmapSegment = Texture_getBgmapSegment(this);
	int pallet = Texture_getPallet(this) << 14;
	
	int charLocation = (CharGroup_getCharSet(this->charGroup) << 9) + CharGroup_getOffset(this->charGroup);
	int i = this->rows;

	//put the map into memory calculating the number of char for each reference
	for(; i--;){
	
		//write into the specified bgmap segment plus the offset defined in the this structure, the this definition
		//specifying the char displacement inside the char mem
		addMem ((void*)BGMap(bgmapSegment) + ((this->xOffset +  (this->yOffset << 6) + (i << 6)) << 1),
				(const u16*)(this->bgmapDef + ( i * (this->cols) << 1)),
				this->cols,
				(pallet) | (charLocation));
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// write an animated and shared map
static void Texture_writeAnimatedShared(Texture this){

	int bgmapSegment = Texture_getBgmapSegment(this);
	int pallet = Texture_getPallet(this) << 14;
	
	// determine the number of frames the map had
	int area = (this->cols * this->rows);
	int charLocation = (CharGroup_getCharSet(this->charGroup) << 9) + CharGroup_getOffset(this->charGroup);
	int frames = CharGroup_getNumberOfChars(this->charGroup) / area;
	
	int i = this->rows;

	//put the map into memory calculating the number of char for each reference
	for(; i--;){
	
		int j = 1;
		//write into the specified bgmap segment plus the offset defined in the this structure, the this definition
		//specifying the char displacement inside the char mem
		for(; j <= frames; j++){
			addMem ((void*)BGMap(bgmapSegment) + ((this->xOffset + (this->cols * (j - 1)) + (this->yOffset << 6) + (i << 6)) << 1),
					(const u16*)(this->bgmapDef + ( i * (this->cols) << 1)),
					this->cols,
					(pallet) | (charLocation + area * (j - 1)));
		}
	}
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// write into memory the chars and this
void Texture_write(Texture this){
	
	//write char group	
	CharGroup_write(this->charGroup);
	
	//if not allocated
	if(0xFF == this->xOffset){
		
		//determine the allocation type
		switch(CharGroup_getAllocationType(this->charGroup)){
		
			case __ANIMATED:
				
				//allocate map in a bgmap segment		
				TextureManager_allocateDedicated(TextureManager_getInstance(), this);
				
				// put the definition in graphic memory
				Texture_writeAnimated(this);
				
				break;
				
			case __ANIMATED_SHARED:
				
				//allocate map in a bgmap segment
				if(TextureManager_allocateShared(TextureManager_getInstance(), this)){

					// put the definitionin graphic memory
					Texture_writeAnimatedShared(this);											
				}
				
				break;
				
			case __NO_ANIMATED:
				
				
				//allocate map in a bgmap segment
				if(TextureManager_allocateShared(TextureManager_getInstance(), this)){
						
					// put the definition in graphic memory
					Texture_writeNoAnimated(this);					
				}
				
				break;

		}
	}		
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// write into memory the chars and this
void Texture_rewrite(Texture this){

	// determinet the allocation type
	switch(CharGroup_getAllocationType(this->charGroup)){
	
		case __ANIMATED:		
			// put the definition in graphic memory
			Texture_writeAnimated(this);
			
			break;
			
		case __ANIMATED_SHARED:
			// put the definitionin graphic memory
			Texture_writeAnimatedShared(this);											

			break;
			
		case __NO_ANIMATED:
			// put the definition in graphic memory
			Texture_writeNoAnimated(this);
			
			break;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// this reallocate a write the bgmap definition in graphical memory
void Texture_resetMemoryState(Texture this){
	
	//fake char offset so it is allocated again
	CharGroup_setOffset(this->charGroup, 0xFF);
	
	//fake bgmap offset so it is allocated again
	this->xOffset = 0xFF;	
	
	// write it in graphical memory
	Texture_write(this);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// write map in hbias mode
void Texture_writeHBiasMode(Texture this){

	// TODO
	/*
	int i;	
	//put the this into memory calculatin the number of char for each reference
	for(i=0;i<this->rows;i++){
		//write into the specified bgmap segment plus the offset defined in the this structure, the this definition
		//specifying the char displacement inside the char mem
		//addMem ((void*)BGTexture(this->bgmapSegment)+((this->xOffset+this->cols/3+(this->yOffset<<6)+(i<<6))<<1), this->bgmapDef+(i<<7), (this->cols/3)*2,(this->pallet<<14)|((CharGroup_getCharSet(&this->charGroup)<<9)+CharGroup_getOffset(&this->charGroup)));
		addMem ((void*)BGTexture(this->bgmapSegment)+((this->xOffset+this->cols/3+64* const this->yOffset+64*i)<<1), this->bgmapDef+(i<<7), (this->cols/3)*2,(this->pallet<<14)|((CharGroup_getCharSet(&this->charGroup)<<9)+CharGroup_getOffset(&this->charGroup)));
	}
	*/	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get texture's number of chars
int Texture_getNumberOfChars(Texture this){
	
	return CharGroup_getNumberOfChars(this->charGroup);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get texture's y offset within bgmap mem
int Texture_getYOffset(Texture this){
	
	return this->yOffset;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get texture's x offset within bgmap mem
int Texture_getXOffset(Texture this){
	
	return this->xOffset;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set texture's y offset within bgmap mem
void Texture_setYOffset(Texture this, int yOffset){
	
	this->yOffset = yOffset;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set texture's x offset within bgmap mem
void Texture_setXOffset(Texture this, int xOffset){
	
	this->xOffset = xOffset;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set texture's rows
void Texture_setRows(Texture this, int rows){
	
	this->rows = rows;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get texture's cols
int Texture_getTotalCols(Texture this){
	
	
	// determine the allocation type
	switch(CharGroup_getAllocationType(this->charGroup)){
	
		case __ANIMATED:
			
			// just return the cols
			return this->cols;
			break;
			
		case __ANIMATED_SHARED:
			
			// return the total number of chars
			return CharGroup_getNumberOfChars(this->charGroup) / this->rows;
			break;
			
		case __NO_ANIMATED:
			
			// just return the cols
			return this->cols;
			break;			
	}
	
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//get texture's rows
int Texture_getTotalRows(Texture this){
	
	return this->rows;
	/*
	switch(CharGroup_getAllocationType(&this->charGroup)){
		case __ANIMATED:
			return this->rows;
			break;
		case __ANIMATEDSHARED:
			return this->rows*CharGroup_getNumberOfChars(&this->charGroup)/this->cols;
			break;
		case __NOANIMATED:
			return this->cols;
			break;			
	}
	return 0;
	*/
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//set texture's cols
void Texture_setCols(Texture this, int cols){
	
	this->cols = cols;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//set texture's bgmap segment
void Texture_setBgmapSegment(Texture this, int bgmapSegment){
	
	this->bgmapSegment = bgmapSegment;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//get texture's bgmap segment
int Texture_getBgmapSegment(Texture this){
	
	return this->bgmapSegment;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//get texture's chargroup
CharGroup Texture_getCharGroup(Texture this){
	
	return this->charGroup;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//set texture's number of chars
void Texture_setNumberOfChars(Texture this, int numberOfChars){

	CharGroup_setNumberOfChars(this->charGroup, numberOfChars);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//set texture's bgmap definition
void Texture_setBgmapDef(Texture this, void *bgmapDef){
	
	this->bgmapDef = bgmapDef;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//get texture's bgmap definition
BYTE* Texture_getBgmapDef(Texture this){
	
	return this->bgmapDef;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set the pallet
void Texture_setPallet(Texture this, int pallet){
	
	this->pallet = pallet;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
int Texture_getPallet(Texture this){
	
	return this->pallet;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve texture's rows
int Texture_getRows(Texture this){
	
	return this->rows;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve texture's cols
int Texture_getCols(Texture this){
	
	return this->cols;
}

