/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												INCLUDES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

//#include <libgccvb2/libgccvb.h>
#include "VirtualList.h"
#include "Constants.h"
#include "MemoryPool.h"
#include "Error.h"


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// class's allocator
VirtualNode VirtualNode_new(const BYTE* const data);

// class's constructor
static void VirtualNode_constructor(VirtualNode this, const BYTE* const data);


// class's destructor
static void VirtualNode_destructor(VirtualNode this);

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#define VirtualNode_ATTRIBUTES				\
											\
	/* it is derivated from*/				\
	Object_ATTRIBUTES						\
											\
	/* pointer to next node */				\
	VirtualNode next;						\
											\
	/* pointer to previos node */			\
	VirtualNode previous;					\
											\
	/* pointer to the data */				\
	BYTE* data;


__CLASS_DEFINITION(VirtualNode, Object);



/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

__CLASS_NEW_DEFINITION(VirtualNode, __PARAMETERS(const BYTE* const data))
__CLASS_NEW_END(VirtualNode, __ARGUMENTS(this, data));

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's constructor
static void VirtualNode_constructor(VirtualNode this, const BYTE* const data){
	
	__CONSTRUCT_BASE(Object,__ARGUMENTS(this));
	
	{
		
		this->next = NULL;
		this->previous = NULL;	

		//make pointer point to data
		this->data = (BYTE*) data;
	}
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
static void VirtualNode_destructor(VirtualNode this){

	// destroy the super object
	__DEALLOCATE(Object);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get class's size
int VirtualNode_getClassSize(){
	return sizeof(VirtualNode_str);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set node's data pointer
void VirtualNode_setData(VirtualNode this, BYTE* data){
	
	this->data = data;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve data pointer
BYTE* VirtualNode_getData(VirtualNode this){
	
	return this->data;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get next node's address
VirtualNode VirtualNode_getNext(VirtualNode this){
	return this->next;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set node's next node's address
void VirtualNode_setNext(VirtualNode this, VirtualNode next){
	
	this->next = next;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set node's previous node's address
VirtualNode VirtualNode_getPrevious(VirtualNode this){
	
	return this->previous;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set node's previous node's address
void VirtualNode_setPrevious(VirtualNode this, VirtualNode previous){
	this->previous = previous;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// swap the data between two nodes
void VirtualNode_swapData(VirtualNode this, VirtualNode node){
	
	VirtualNode_str auxNode;
	// check that both nodes are valid and are not the same
	if(!(this && node && (this != node))){
		return;
	}
	
	// save this's attributes
	auxNode.next = this->next;
	auxNode.previous = this->previous;
	
	// swap next and previous pointers
	if(this->previous){
		this->previous->next = node;
	}
	if(node->next){
		node->next->previous = this;
	}

	this->next = node->next;
	this->previous = node;
	
	node->next = this;
	node->previous = auxNode.previous;
}







/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											 CLASS'S MACROS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// define a limit to prevent, and detect looped lists
#define LISTMAXSIZE 1000


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */
#define VirtualList_ATTRIBUTES					\
												\
	/* it is derivated from*/					\
	Object_ATTRIBUTES							\
												\
	/* a pointer to the head of the list */ 	\
	VirtualNode head;							\
												\
	/* a pointer to the tail of the list */		\
	VirtualNode tail;



__CLASS_DEFINITION(VirtualList, Object);

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// class's constructor
static void VirtualList_constructor(VirtualList this);


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

__CLASS_NEW_DEFINITION(VirtualList, __PARAMETERS())
__CLASS_NEW_END(VirtualList, __ARGUMENTS(this));


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's constructor
static void VirtualList_constructor(VirtualList this){

	__CONSTRUCT_BASE(Object,__ARGUMENTS(this));	

	{
		// construct it
		this->head = NULL;		
		this->tail  = NULL;
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
void VirtualList_destructor(VirtualList this){

	// make sure we remove all nodes
	VirtualList_clear(this);
	
	__DEALLOCATE(Object);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// remove all nodes from the list
void VirtualList_clear(VirtualList this){

	if(this->head){
		
		// point to the head
		VirtualNode node = this->head;
		
		// move the head to next node
		this->head = this->head->next;
		
		// while there are nodes
		while(node){
			
			// call destructor
			__DELETE(node);
			
			//move the node to the head
			node = this->head;
			
			//move the head
			if(this->head){

				this->head = this->head->next;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// add a new node to the begging of the list
int VirtualList_pushFront(VirtualList this, const BYTE* const data){
	
	// create a node
	VirtualNode newNode = __NEW(VirtualNode, __ARGUMENTS(data));
	
	//set previous if list isn't empty
	if(this->head){
		
		this->head->previous = newNode;
	}
	
	// assign the node to the head of the list	
	newNode->next = this->head;
	
	//move the head
	this->head = newNode;

	// set the tail
	if(!this->tail){
		
		this->tail = this->head;
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// remove the first element in the list
void VirtualList_popFront(VirtualList this){

	//if head isn't null
	if(this->head){

		VirtualNode node = this->head;
		
		if(node->next){
			
			this->head = node->next;
			
			//move head's previous pointer
			this->head->previous=NULL;
		}
		else{

			//set head
			this->head = NULL;
		}			

		//free dynamic memory
		__DELETE(node);		
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// add a new node to the end of the list
// TODO: remove method does not take into account the tail
// don't use this method!
static VirtualNode VirtualList_pushBack(VirtualList this, const BYTE* const data){

	VirtualNode newNode = __NEW(VirtualNode, __ARGUMENTS(data));
	
	ASSERT(data);
	
	if(!newNode){
		
		ASSERT(false);
		return false;
	}
	
	// set previous if list isn't empty
	if(this->tail){
		
		this->tail->next = newNode;
	}
	
	// assign the node to the head of the list	
	newNode->previous = this->tail;
	
	//move the tail
	this->tail = newNode;
	
	// set the tail
	if(!this->head){
		
		this->head = this->tail;
	}
	
	return newNode;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve the number of objects the list has
int VirtualList_getSize(VirtualList this){

	int counter = 0;
	
	VirtualNode node = this->head;
	
	while(node){
		
		//load next node
		node = node->next;
		
		++counter;
		
		//increment counter	
		ASSERT(counter < LISTMAXSIZE, VL_ENDLESS_GSIZE_ERR);
		
	}
	return counter;

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// return data pointer of object in the given index node
void* VirtualList_getNodeData(VirtualList this, int item){
	
	// get the node
	VirtualNode node = VirtualList_getNode(this, item);
	// return the data
	return (node)? node->data: NULL;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get node at item position
VirtualNode VirtualList_getNode(VirtualList this, int item){
	
	int counter=0;
	//get list's size
	
	int listSize = VirtualList_getSize(this);
	//load head
	
	VirtualNode node = this->head;
	
	//if not null head
	if(node){
		
		//if item hasn't reached list's size
		if(item < listSize){
			
			//increase counter while node hasn't reached list's end
			//and counter hasn't reached the item requested
			while((node) && (counter < item)){
				
				//increase counter
				counter++;
				
				//load next node
				node = node->next;
				
				//if item reached
				if(counter == item){
					//return node's data
					return node;
				}
			}
			//if item reached
			if(counter == item){
				//return node's data
				return node;
			}
			return NULL;
		}
	}
	return NULL;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get node's address of node containing datapointer
void* VirtualList_getObject(VirtualList this, void* const dataPointer){
	
	VirtualNode node = this->head;
	
	// check if data pointer is valid
	if(!dataPointer){
		return NULL;
	}
	
	//locate node
	while(node && node->data != dataPointer){
		node = node->next;
	}
	
	return node;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// remove a node
static int VirtualList_removeNode(VirtualList this, VirtualNode node){

	//if node isn't null
	if(node){

		// if the node is the head of the list
		if(node == this->head){
			
			if(node->next){
				this->head=node->next;
				//move head's previous pointer
				this->head->previous=NULL;
			}
			else{

				//set head
				this->head = NULL;
			}			
		}
		else{
			
			//if node isn't the last in the list
			if(node->next){
				
				//join the previous and next nodes
				node->next->previous = node->previous;
				node->previous->next = node->next;
				
			}
			else{
				
				//set end of list
				node->previous->next = NULL;
			}
			
		}		

		//free dynamic memory
		__DELETE(node);		


		return true;
		
	}

	ASSERT(false, "VirtualList: removing empty node");
	
	return false;

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// remove a node from the list
int VirtualList_removeElement(VirtualList this, const void* const dataPointer){
	
#ifdef __DEBUG
	int counter = 0;
#endif
	
	VirtualNode node = this->head;
	

	//locate node
	while(node && node->data != dataPointer){
		
		node = node->next;		
		ASSERT(++counter < LISTMAXSIZE, VL_ENDLESS_REM_ERR);
	}
	
	ASSERT(node, VirtualList: node not found);

	return VirtualList_removeNode(this, node);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// copy source list's elements to destiny list
void VirtualList_copy(VirtualList this, VirtualList sourceList){
	
#ifdef __DEBUG
	int counter = 0;
#endif
	
	VirtualNode node = sourceList->head;
	
	VirtualList_clear(this);
	
	while(node){
		//add next node 
		VirtualList_pushBack(this, node->data);
		//move to next node
		node = node->next;
		
		ASSERT(++counter < LISTMAXSIZE, VL_ENDLESS_COPY_ERR);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve list's head's address
VirtualNode VirtualList_begin(VirtualList this){
	return this->head;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve the first element
BYTE* const VirtualList_front(VirtualList this){
	
	return this->head->data;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve last list's node
VirtualNode VirtualList_end(VirtualList this){

	return this->tail;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve the last element
BYTE* const VirtualList_back(VirtualList this){
	
	return this->tail->data;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// insert a node after the node specified
VirtualNode VirtualList_insertAfter(VirtualList this, VirtualNode node, const BYTE* const data){


	if(node == this->head){
		VirtualList_pushBack(this, data);
		return NULL;
	}
	{
		VirtualNode newNode = __NEW(VirtualNode, __ARGUMENTS(data));
		
		if(!newNode){
			return false;
		}
	
		//set previous if list isn't empty
		newNode->next = node->next;
		
		if(node->next){
			node->next->previous = newNode;
		}
		node->next = newNode;
		newNode->previous = node;
		return newNode;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// insert a node before the node specified
VirtualNode VirtualList_insertBefore(VirtualList this, VirtualNode node, const BYTE* const data){
	
	if(node == this->head){
		VirtualList_pushBack(this, data);
		return NULL;
	}
	{
		VirtualNode newNode = __NEW(VirtualNode, __ARGUMENTS(data));
		
		if(!newNode){
			return false;
		}

		//set previous if list isn't empty	
		newNode->next = node;
		newNode->previous = node->previous;
		node->previous->next = newNode;
		node->previous = newNode;
		return newNode;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// swap two lists heads and tails
void VirtualList_swap(VirtualList this, VirtualList secondList){
	
	// swap heads
	VirtualNode aux = this->head;
	
	this->head = secondList->head;
	
	secondList->head = aux;
	
	// swap tails
	aux = this->tail;
		
	this->tail = secondList->tail;
		
	secondList->tail = aux;
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get node's address at given position
BYTE* VirtualList_getObjectAtPosition(VirtualList this, int position){
	
	VirtualNode node = this->head;
	
	int counter = 0;
	
	if(position < 0){
		return NULL;
	}
	//locate node
	for(; node && counter < position; node = node->next, counter++);
	

	if(counter < VirtualList_getSize(this)){
		return node->data;
	}
	
	return NULL;
}