/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */


/*---------------------------------INCLUDES--------------------------------*/
#include "Mario.h"
#include "MarioIdle.h"
#include "MarioWin.h"
#include "MarioStunned.h"
#include "MarioDead.h"
#include "MarioOnBridge.h"
#include <config.h>
#include "../koopa/KoopaStunned.h"
#include "../koopa/Koopa.h"
#include "../../objects.h"

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											 CLASS'S MACROS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// Mario.c

// A Mario!
__CLASS_DEFINITION(Mario, Actor); 

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


// a static member of this class
static u32 gameLayers[TOTAL_GAME_LAYERS] = {
	ITOFIX23_9(LAYER_0),
	ITOFIX23_9(LAYER_1),
	ITOFIX23_9(LAYER_2),
	ITOFIX23_9(LAYER_3),
	ITOFIX23_9(LAYER_4),
};



/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

// Only one instance
Mario mario = NULL;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// there can only be one mario instantiated
Mario Mario_getInstance(){
	
	return mario;
}

void Mario_setInstance(Mario instance){
	
	ASSERT(!mario, Mario already instantiated);
	
	mario = instance;
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// always call these to macros next to each other
__CLASS_NEW_DEFINITION(Mario, __PARAMETERS(ActorDefinition* actorDefinition, int ID))
__CLASS_NEW_END(Mario, __ARGUMENTS(this, actorDefinition, ID));

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's constructor
void Mario_constructor(Mario this, ActorDefinition* actorDefinition, int ID){
	
	__CONSTRUCT_BASE(Actor, __ARGUMENTS(this, actorDefinition, ID));
	
	this->energy = 1;

	// initialize me as idle
	StateMachine_swapState(this->stateMachine, (State)MarioIdle_getInstance());

	// register a shape for collision detection
	this->shape = CollisionManager_registerShape(CollisionManager_getInstance(), (InGameEntity)this, kRect, actorDefinition->inGameEntityDefinition.deep);

	// I'm the focus actor
	Stage_setFocusActor(Game_getStage(Game_getInstance()), (Actor)this);
	
	// I'm not holding anything
	this->holdObject = NULL;
	
	// I alwasy start in the first layer
	this->layer = 0;
	
	this->movingOverZ = false;
	
	this->actionTime = 0;
	
	this->bridge = NULL;
	
	Mario_setInstance(this);
	
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
void Mario_destructor(Mario this){

	// free the instance pointer
	ASSERT(mario == this, Mario: more than on instance);
	mario = NULL;

	// delete the super object
	__DEALLOCATE(Actor);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// start movement
void Mario_startMoving(Mario this){

	// must do some comparison regardless of my current direction
	int  compareDirection = 0;
	int  newDirection = 0;

	// pick the current pressed key
	u16 currentPressedKey = vbKeyPressed();
		
	// determine which direction will compare
	if(currentPressedKey & K_LL){
		
		compareDirection = __RIGHT;
		newDirection = __LEFT;
	}
	else{

		if(currentPressedKey & K_LR){
			compareDirection = __LEFT;
			newDirection = __RIGHT;
			
		}
	}
	
	// if a direction was determined 
	if(newDirection){
		
		this->movingOverZ = false;
		
		// begin to move
		Actor_startXMovement((Actor)this, __RETARMOVEX, MARIO_VELOCITY_X, MARIO_ACCELERATION_X);
		
		// must change the direction ?
		if(this->direction.x == compareDirection){				
			
			this->direction.x = newDirection;
		}
	}		

	// only play walk animation if not jumping/falling
	if(!this->movementState.y){
		
		if(this->holdObject){
			
			Actor_playAnimation((Actor)this, "WalkHolding");
		}
		else{
			
			Actor_playAnimation((Actor)this, "Walk");
		}
	}		
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// update movement
void Mario_move(Mario this){

	// pick the current pressed key
	u16 currentPressedKey = vbKeyPressed();
	
	// check if must throw the object
	if(this->holdObject){
	
		Mario_throwObject(this);
	}

	if(this->movementState.z){

		if(__FAR == this->direction.z && this->position.z > gameLayers[this->layer]){

			Actor_stopZMovement((Actor)this);
			
			this->position.z = gameLayers[this->layer];
			
			return;
		}
		else{
			
			if(__NEAR == this->direction.z && this->position.z < gameLayers[this->layer]){
				
				Actor_stopZMovement((Actor)this);
				
				this->position.z = gameLayers[this->layer];
				
				return;
			}
		}

		
		// move the actor
		Actor_move((Actor)this);
		
		// if I'm holding something
		if(this->holdObject){
			
			// update object's position
			Mario_updateHoldObjectPosition(this);
		}		
		return;
	}
	
	//if mario is moving and all keys are left
	if(!currentPressedKey  || (!(currentPressedKey & K_LR) && !(currentPressedKey &  K_LL))){

		if(!(this->moveType &__RETARMOVEX)){

			// must start to stop
			this->moveType |=__RETARMOVEX;
			
			// don't slide if holding something 
			if(!this->holdObject){
				
				Actor_playAnimation((Actor)this, "Slide");
			}
			
			this->velocity.x = MARIO_VELOCITY_X;
		}
	}
	else{

		char walkStr[] = "Walk";
		char walkHoldStr[] = "WalkHolding";
		char* walkAnimation = walkStr;
		
		if(this->holdObject){
			
			walkAnimation = walkHoldStr;
		}

		// if it is only moving over the x axis and walking animation is no playing
		if(!this->movementState.y && !AnimatedSprite_isPlayingFunction((AnimatedSprite)this->sprite, this->actorDefinition->animationDescription, walkAnimation)){
			
			// must play walk animation of not playing it
			Actor_playAnimation((Actor)this, walkAnimation);
			
			Actor_startXMovement((Actor)this, __RETARMOVEX, MARIO_VELOCITY_X, MARIO_ACCELERATION_X);
		}
		
	}

	// if movement is uniform
	if(!(this->moveType &__RETARMOVEX)){

		// if B button is pressed
		if(currentPressedKey & K_B){
			
			// boost velocity
			this->velocity.x = FIX23_9_MULT(MARIO_VELOCITY_X, MARIO_SPEED_MULTIPLIER_X);
		}
		else{
			
			// shoot the shell
			this->velocity.x = MARIO_VELOCITY_X;
		}
	}

	// might be jumping
	// TODO: increse y velocity if A button hold while jumping
	if(this->movementState.y){
		
		if(__UP == this->direction.y){
			
			if(vbPadPreviousKey() & K_A){
				
				//this->acceleration.y = (MARIO_ACCELERATION_Y + __GRAVITY) * (-1) / MARIO_SPEED_MULTIPLIER_X;
			}
		}
	}


	// if not jumping over z
	if(!this->movingOverZ){

		// if mario is falling
		if(this->movementState.y && __DOWN == this->direction.y ) {
	
			char fallStr[] = "Fall";
			char fallHoldStr[] = "FallHolding";
			char* fallAnimation = fallStr;
			
			if(this->holdObject){
				
				fallAnimation = fallHoldStr;
			}
	
			// and not playing corresponding animation
			if(!AnimatedSprite_isPlayingFunction((AnimatedSprite)this->sprite, this->actorDefinition->animationDescription, fallAnimation)){
			
				// play animation
				Actor_playAnimation((Actor)this, fallAnimation);
			}
		}
	}
	
	// move the actor
	Actor_move((Actor)this);

	// if I'm holding something
	if(this->holdObject){
		
		// update object's position
		Mario_updateHoldObjectPosition(this);
	}

	//if(kDead != Actor_getInGameState((Actor)owner) && !(ownerMovementState.x || ownerMovementState.y || ownerMovementState.z)){
	if(this->inGameState != kDead && !(this->movementState.x | this->movementState.y | this->movementState.z)){
		
		// go back to idle
		StateMachine_swapState(this->stateMachine, (State)MarioIdle_getInstance());
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// make him jump
void Mario_jump(Mario this){

	// if not jumping/falling
	if(this->movementState.y == __PASSIVE && vbNewKeyPressed(K_A)){

		// jump!
		Actor_jump((Actor)this, MARIO_VELOCITY_Y, MARIO_ACCELERATION_Y);
		SoundManager_loadFxSound(SoundManager_getInstance(), (u16*)MARIO_JUMP_SND, this->position);
		
		// check if must jump over z axis
		if(!Mario_checkIfZJump(this)){				
			
			if(this->holdObject){

				// play animation
				Actor_playAnimation((Actor)this, "JumpHolding");
			}
			else{
				
				// play animation
				Actor_playAnimation((Actor)this, "Jump");
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// check direction
void Mario_checkDirection(Mario this, u16 currentPressedKey){

	u16 compareKey = 0;
	int  newDirection = 0;
	
	// otherwise I was already moving		
	Size worldSize = Stage_getSize(Game_getStage(Game_getInstance()));
	
	// movement is uniform
	this->moveType &= ~(__ACCELMOVEX | __RETARMOVEX);

	// check if player wants to move to the oposite direction
	switch(this->direction.x){
	
		case __LEFT:
	
			// don't go below x = 0
			if(this->position.x < ITOFIX23_9(Entity_getWidth((Entity)this))){
				
				Actor_stopXMovement((Actor)this);
				
				this->position.x = ITOFIX23_9(Entity_getWidth((Entity)this));
			}
			
			// check if player has pressed right
			compareKey = K_LR;
			
			newDirection = __RIGHT;
			break;
			
		case __RIGHT:
			
			// don't go above x = wordSize.x
			if(this->position.x > ITOFIX23_9(worldSize.x) - ITOFIX23_9(Entity_getWidth((Entity)this))){
				
				Actor_stopXMovement((Actor)this);
			}

			// check if player has pressed left
			compareKey = K_LL;
			
			newDirection = __LEFT;
			break;					
	}

	// if the current pressed key is the same as the one to compare
	if((currentPressedKey & compareKey)
			&&
			!((K_LR & currentPressedKey) && (K_LL & currentPressedKey))
			){					

		// change direction
		this->direction.x = newDirection;				
		
		if(!this->movingOverZ){
			
			// start moving in the new direction
			Actor_startXMovement((Actor)this, __RETARMOVEX, MARIO_VELOCITY_X, MARIO_ACCELERATION_X);
		}
	}

	// don't allow movement over x while moving over z
	if(this->movingOverZ){
		
		return;
	}
	
	if(!this->movementState.x && !this->movingOverZ){
		
		// start moving in the new direction
		Actor_startXMovement((Actor)this, __RETARMOVEX, MARIO_VELOCITY_X, MARIO_ACCELERATION_X);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// process a collision
int  Mario_processCollision(Mario this, Telegram telegram){
	
	// retrieve message
	int message = Telegram_getMessage(telegram);
	
	// retrieve collision entity
	InGameEntity inGameEntity = (InGameEntity) Telegram_getExtraInfo(telegram);
	
	// process message
	switch(message){

		// if it is a collision
		case kCollisionXY:
		case kCollisionX:
		case kCollisionY:
		case kCollisionZ:			

			// determine the type of colliding entity
			switch(InGameEntity_getInGameType(inGameEntity)){
			
				case kGate:
					
					if(__PASSIVE == this->movementState.z){
						
						Actor_playAnimation((Actor)this, "Win");
						Actor_stopMovement((Actor)this);
						
					}
					return true;
					break;
				case kGoal:
					
					if(!StateMachine_isInState(this->stateMachine, (State)MarioWin_getInstance())){
						// if i have something being hold
						if(this->holdObject){
						
							// hold object must be destroyed
							__VIRTUAL_CALL(void, Actor, die, (Actor)this->holdObject);
		
						}
						
						// change state
						StateMachine_swapState(this->stateMachine, (State)MarioWin_getInstance());
					}
					break;
				// if solid
				case kSolid:

					//if moving over the bridge
					if(this->movementState.z && this->bridge){
						
						return false;
					}

					// stop movement over colliding axis and align to object if needed
					if(this->movementState.z){

						// align to the colliding object
						Actor_alignZ((Actor)this, (InGameEntity)inGameEntity, 1);
						
						// finally stop the movement
						Actor_stopZMovement((Actor)this);
						return true;
					}

					// test x axis
					if(this->movementState.x && (kCollisionX == message || kCollisionXY == message)){
						
						// align to the colliding object
						Actor_alignX((Actor)this, (InGameEntity)inGameEntity, 1);
						
						// finally stop the movement
						Actor_stopXMovement((Actor)this);
						
						return true;
					}

					// test y axis
					if(this->movementState.y && (kCollisionY == message || kCollisionXY == message)){

						// align to the colliding object
						Actor_alignY((Actor)this, (InGameEntity)inGameEntity, 1);
						
						// finally stop the movement
						Actor_stopYMovement((Actor)this);
						
						// save object below
						if(this->position.y < Entity_getPosition((Entity)inGameEntity).y){
							
							// save object below me
							this->objectBelow = inGameEntity;
							
						}						
						return true;
					}

					return false;
					break;
					
				case kBridge:

					// if I'm moving over the bridge
					if(this->bridge != (Bridge)Telegram_getExtraInfo(telegram) && StateMachine_isInState(this->stateMachine, (State)MarioOnBridge_getInstance())){
					
						VBVec3D bridgePosition = Entity_getPosition((Entity)Telegram_getExtraInfo(telegram));
						
						StateMachine_swapState(this->stateMachine, (State)MarioIdle_getInstance());
						
						// place myself behind or infront of the bridge
						if(__NEAR == Bridge_getDirection(this->bridge)){
							
							this->position.z = bridgePosition.z - ITOFIX23_9(1);
							
						}
						else{
							
							this->position.z = bridgePosition.z + ITOFIX23_9(1);
						}
						
						this->position.y = bridgePosition.y +
						                   ITOFIX23_9(
						                		   Entity_getHeight((Entity)this->bridge) / 2
						                		   - Entity_getHeight((Entity)this) / 2	- 1);
						
						// determine which layer I'm
						Mario_determineLayer(this);
						

						Actor_stopMovement((Actor)this);
						
					}
					
					// register bridge
					this->bridge = (Bridge)Telegram_getExtraInfo(telegram);
					
					// update hold object's position
					Mario_updateHoldObjectPosition(this);

					return false;
					break;

				case kPiranhaPlant:
					
					// i just take the hit
					Mario_takeHit(this, Entity_getPosition((Entity)inGameEntity));
					break;

				// if it was a koopa
				case kKoopa:
					{

						// determine if the koopa hits mario						
						if(Mario_isHitByEnemy(this, (Enemy)Telegram_getExtraInfo(telegram), message)){

							// take a hit
							Mario_takeHit(this, Entity_getPosition((Entity)inGameEntity));
							
							// tell koopa to attack
							Koopa_attack((Koopa)inGameEntity);
						}
						
						// we want to process solid object
						return true;
					}					
					break;
			}
			
			break;
	}
	return false;

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Mario_takeHit(Mario this, VBVec3D position){
	
	// first stop all movement
	Actor_stopMovement((Actor)this);
	
	// reduce energy
	if(--this->energy){

	}
	else{

		// not die
		Mario_die(this);
	}
	
	// determine which animation to play 
	// based on my position relative to the
	// enemy position
	if(position.x > this->position.x){
		
		if(__RIGHT == this->direction.x){
			
			Actor_playAnimation((Actor)this, "HitFront");
		}
		else{
		
			Actor_playAnimation((Actor)this, "HitBehind");
		}
	}
	else{
		if(__LEFT == this->direction.x){
			
			Actor_playAnimation((Actor)this, "HitFront");
		}
		else{
		
			Actor_playAnimation((Actor)this, "HitBehind");
		}							
	}
	
	// register time
	this->actionTime = Clock_getTime(_inGameClock);
	
	// must unregister the shape for collision detections
	Shape_setActive(this->shape, false);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
int Mario_isHitByEnemy(Mario this, Enemy enemy, int axis){

	ASSERT(enemy);

	
	// if enemy is already dead
	if(kDead == Actor_getInGameState((Actor)enemy)){
		
		return false;
	}
	
	// if it is the holdobject which wants to hit me
	if(this->holdObject == (Actor)enemy){
		
		this->holdObject = NULL;
		return true;
	}
	

	// check if I'm over the koopa
	if((this->movementState.y && __DOWN == this->direction.y) && 
	    (
    		this->position.y + ITOFIX23_9(Entity_getHeight((Entity)this) >> 2) <
    		Entity_getPosition((Entity)enemy).y
    	)
    ){
		
		// align to the colliding object
		Actor_alignY((Actor)this, (InGameEntity)enemy, 2);
		
		//check if player wants to jump over z axis
		if(!Mario_checkIfZJump(this)){
			
			if(this->holdObject){

				// play animation
				Actor_playAnimation((Actor)this, "JumpHolding");
			}
			else{
				
				// play animation
				Actor_playAnimation((Actor)this, "Jump");
			}
		}

		// if A button is pressed boost the y velocity
		if(K_A & vbKeyPressed()){

			// I will bounce
			Actor_jump((Actor)this, FIX23_9_MULT(MARIO_VELOCITY_Y, MARIO_SPEED_MULTIPLIER_Y), MARIO_ACCELERATION_Y);
		}
		else
		{
			// I will bounce
			Actor_jump((Actor)this, MARIO_VELOCITY_Y , MARIO_ACCELERATION_Y);
		}
		

		// tell koopa I've hit him
		if(this->movingOverZ){

			__VIRTUAL_CALL(void, Enemy, takeHit, (Enemy)enemy, __VIRTUAL_ARGUMENTS(__ZAXIS, this->direction.z));
		}
		else{
			
			__VIRTUAL_CALL(void, Enemy, takeHit, (Enemy)enemy, __VIRTUAL_ARGUMENTS(__XAXIS, this->direction.x));
		}
			
		return false;
	}

	// if I'm holding something and facing to the side
	if(this->holdObject && !this->movingOverZ){
		
		
		// if I'm facing the enemy
		if((this->position.x <  Entity_getPosition((Entity)enemy).x && __RIGHT == this->direction.x)
			||
			(this->position.x >  Entity_getPosition((Entity)enemy).x && __LEFT == this->direction.x)
			){
			// tell enemy to die
			__VIRTUAL_CALL(void, Actor, die, (Actor)enemy);
			
			// hold object must be destroyed
			__VIRTUAL_CALL(void, Actor, die, (Actor)this->holdObject);
			
			// holding nothing now
			this->holdObject = NULL;
			
			return false;
		}
	}

	// if collision was over x axis
	if(kCollisionX == axis && !(__VIRTUAL_CALL(int, Enemy, canAttack, enemy) ||	this->holdObject)){

		// if B button is pressed, pick up 
		if(K_B & vbKeyPressed()){
			
			if(!this->holdObject){
				
				Mario_pickupObject(this, (Actor)enemy);
				
				return false;
			}
		}
		
		// stop my movement
		Actor_stopXMovement((Actor)this);

		// align to the colliding object
		Actor_alignX((Actor)this, (InGameEntity)enemy, 1);

		// tell koopa to begin bouncing
		__VIRTUAL_CALL(void, Enemy, takeHit, (Enemy)enemy, __VIRTUAL_ARGUMENTS(__XAXIS, this->direction.x));

		// I will bounce
		Actor_jump((Actor)this, FIX23_9_DIV(MARIO_VELOCITY_Y, ITOFIX23_9(2)), MARIO_ACCELERATION_Y);
		
		if(this->holdObject){

			// play animation
			Actor_playAnimation((Actor)this, "JumpHolding");
		}
		else{
			
			// play animation
			Actor_playAnimation((Actor)this, "Jump");
		}
		return false;
	}
	
	// otherwise... I was hit by the enemy
	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set action time
void Mario_setActionTime(Mario this, u32 time){
	
	this->actionTime = time;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve action time
u32 Mario_getActionTime(Mario this){

	return this->actionTime;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// pickup an object
void Mario_pickupObject(Mario this, Actor object){
	
	this->holdObject = object;
	
	// set hold object's position 
	Mario_updateHoldObjectPosition(this);
	
	// remove from collision detection system
	InGameEntity_setShapeState((InGameEntity)this->holdObject, false);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
int Mario_checkIfZJump(Mario this){

	// if up/down pressed
	u16 pressedKey = vbKeyPressed();
	
	this->movingOverZ = false;
	
	// if I'm on a bridge
	if(this->bridge){
	
		// can't jump over z
		return false;
	}
	
	if((pressedKey & K_LU) && this->layer < (TOTAL_GAME_LAYERS -1 )){
	
		// moving to next layer
		this->layer++;
		
		// play animation
		Actor_playAnimation((Actor)this, "JumpBack");
		
		// don't allow movement over x axis
		Actor_stopXMovement((Actor)this);
		
		// I'm going further
		this->direction.z = __FAR;
		
		// start moving in the new direction
		Actor_startZMovement((Actor)this, ~(__ACCELMOVEX | __RETARMOVEX),
				MARIO_VELOCITY_Z, MARIO_ACCELERATION_Z);

		this->movingOverZ = true;

		return true;
	}
	else{
	
		if((pressedKey & K_LD) && this->layer > 0){
		
			// moving to previous layer
			this->layer--;
			
			// play animation
			Actor_playAnimation((Actor)this, "JumpFront");
			
			// don't move over x axis
			Actor_stopXMovement((Actor)this);
			
			// I'm  moving toward the screen
			this->direction.z = __NEAR;
		
			// start movement
			Actor_startZMovement((Actor)this, ~(__ACCELMOVEX | __RETARMOVEX)
					, MARIO_VELOCITY_Z, MARIO_ACCELERATION_Z);

			this->movingOverZ = true;
			
			return true;
		}
	}

	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Mario_fallDead(Mario this){

	Actor_playAnimation((Actor)this, "HitFront");
	
	Mario_die(this);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// called when blink animation is done
void Mario_blinkDone(Mario this){
	
	// play idle animation depende on whether I'm holding something or nor
	if(this->holdObject){
		
		Actor_playAnimation((Actor)this, "IdleHolding");
	}
	else{
		
		Actor_playAnimation((Actor)this, "Idle");
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// was jumping over z?
int  Mario_isMovingOverZ(Mario this){
	
	return this->movingOverZ;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set jumping over z status
void Mario_setMovingOverZ(Mario this, int  state){
	
	this->movingOverZ = state;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set hold object's position 
void Mario_updateHoldObjectPosition(Mario this){
	
	
	// make sure there is an object being hold
	if(!this->holdObject){
		
		return;
	}
	{
		VBVec3D position = Entity_getPosition((Entity)this);
		
	
		if(!this->movingOverZ){
		
			position.x += ITOFIX23_9(MARIO_HOLD_OBJECT_X * this->direction.x);
			position.y += ITOFIX23_9(MARIO_HOLD_OBJECT_Y);
		}
		
		position.z += ITOFIX23_9(MARIO_HOLD_OBJECT_Z * this->direction.z);
		
		Entity_setPosition((Entity)this->holdObject, position);
	}
	{

		MovementState movementState = {this->movementState.x | this->movementState.y, __PASSIVE, this->movingOverZ};

		Actor_setMovementState(this->holdObject, movementState);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// make mario to look to the player
void Mario_lookFront(Mario this){

	// I'm looking to the screen
	this->direction.z = __NEAR;
	
	// play animation
	Actor_playAnimation((Actor)this, "Front");

	// change it to being able to update properly hold object's position
	this->movingOverZ = true;
	
	// update hold object's position
	Mario_updateHoldObjectPosition(this);
	
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// make mario to look away the player
void Mario_lookBack(Mario this){

	// I'm looking away the screen
	this->direction.z = __FAR;

	// play animation
	Actor_playAnimation((Actor)this, "Back");

	// change it to being able to update properly hold object's position
	this->movingOverZ = true;
	
	// update hold object's position
	Mario_updateHoldObjectPosition(this);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set  graphical gap
void Mario_setGap(Mario this){

	this->gap = this->inGameEntityDefinition->gap;
	
	// if I'm not in the edge, return ROM gap
	if(Actor_isPlayingAnimation((Actor)this, "OnEdge")){
	
		// this is texture specific
		this->gap.left += 4;
		this->gap.right += 4;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// make mario to look to the side
void Mario_lookSide(Mario this){
	
	char blinkStr[] = "Blink";
	char idleStr[] = "Idle";
	
	char blinkHoldStr[] = "BlinkHolding";
	char idleHoldStr[] = "IdleHolding";
	
	char* animationToPlay;

	// I'm always going to the screen while looking to the side
	this->direction.z = __NEAR;
	
	// if I'm holding something
	if(this->holdObject){

		animationToPlay = blinkHoldStr; 
	}
	else{
		
		animationToPlay = blinkStr;
	}

	// first check if I'm over something
	if(this->objectBelow){
		
		VBVec3D position = Entity_getPosition((Entity)this->objectBelow);
		int width = ITOFIX23_9(abs(Entity_getWidth((Entity)this->objectBelow) - (Entity_getWidth((Entity)this) >> 1)) >> 1) ;
	
		if(this->position.x > position.x + width || this->position.x < position.x - width){
			
			Actor_playAnimation((Actor)this, "OnEdge");
			
			return;
		}
	}
	
	// if not playing blinking animation
	if(!Actor_isPlayingAnimation((Actor)this, animationToPlay)){

		u32 currentTime = Clock_getTime(_clock);
		
		// randomly select if play blinking
		if(currentTime - this->actionTime > MARIO_BLINK_DELAY){
			
			Actor_playAnimation((Actor)this, animationToPlay);
			
			// back to normal
			this->movingOverZ = false;

			// update hold object's position
			Mario_updateHoldObjectPosition(this);
			
			this->actionTime = currentTime;
		}
		else{
			
			// otherwhise
			if(animationToPlay == blinkHoldStr){
			
				animationToPlay = idleHoldStr;
			}
			else{
				
				animationToPlay = idleStr;
			}
		
			// if not playing idle animation
			if(!Actor_isPlayingAnimation((Actor)this, animationToPlay)){
			
				Actor_playAnimation((Actor)this, animationToPlay);
				
				// back to normal
				this->movingOverZ = false;

				// update hold object's position
				Mario_updateHoldObjectPosition(this);
				
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// check if must thrown an object
void Mario_throwObject(Mario this){
	
	
	// if I'm not holding anything
	if(!this->holdObject){
		
		return;
	}
	
	// if player released B button
	if(!(K_B & vbKeyPressed())){

		s8 direction = this->direction.x;
		int  axis = __XAXIS;
		


		// first place it in the same z coordinate
		//VBVec3D position = Entity_getPosition((Entity)this->holdObject);
		
		Mario_updateHoldObjectPosition(this);
		
		//position.z = this->position.z;
		//Entity_setPosition((Entity)this->holdObject, position);

		// if moving over z axis
		if(this->movingOverZ){
		
			direction = this->direction.z;
			axis = __ZAXIS;
		}
		
		// firt align the object to be thrown 
		// to not produce a collision with me after
		// throwing it
		if(__XAXIS == axis){
			
			Actor_alignX(this->holdObject, (InGameEntity)this, 3);
		}
		else{

			Actor_alignZ(this->holdObject, (InGameEntity)this, this->actorDefinition->inGameEntityDefinition.deep + 1);
		}
		
		// inform the object that it has been thrown
		__VIRTUAL_CALL(void, Actor, beThrown, this->holdObject, __VIRTUAL_ARGUMENTS(axis, direction));
	
		// add to collision detection system again
		InGameEntity_setShapeState((InGameEntity)this->holdObject, true);

		// not holding it anumore
		this->holdObject = NULL;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// die mario
void Mario_die(Mario this){

	// go to dead state
	StateMachine_swapState(this->stateMachine, (State)MarioDead_getInstance());
	
	// if I have something being hold
	if(this->holdObject){
	
		// hold object must be destroyed
		__VIRTUAL_CALL(void, Actor, die, (Actor)this->holdObject);

	}

	// register time
	this->actionTime = Clock_getTime(_inGameClock);
	
	// must unregister the shape for collision detections
	Shape_setActive(this->shape, false);

	// change in game state
	this->inGameState = kDead;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// check if must move over a bridge
void Mario_checkIfBridge(Mario this, int keyPressed){

	
	// if I'm close to a bridge
	if(this->bridge){
		
		char walkFrontStr[] = "WalkFront";
		char walkFrontHoldStr[] = "WalkFront";
		char walkBackStr[] = "WalkBack";
		char walkBackHoldStr[] = "WalkBack";
		
		
		char* walkAnimation = walkFrontStr;

		// if up key pressed
		if(K_LU & keyPressed){
			
			if(__FAR != Bridge_getDirection(this->bridge)){
				
				return;
			}
		}

		// if down key pressed
		if(K_LD & keyPressed){
			
			if(__NEAR != Bridge_getDirection(this->bridge)){
				
				return;
			}
		}

		// check that I'm close to the bridge
		{
			VBVec3D position = Entity_getPosition((Entity)this->bridge);
			int width = ITOFIX23_9(Entity_getWidth((Entity)this->bridge) >> 1);
			Gap gap = InGameEntity_getGap((InGameEntity)this->bridge);
			
			if(this->position.x < position.x - width + ITOFIX23_9(gap.left) 
			   || 
			   this->position.x > position.x + width - ITOFIX23_9(gap.right)){

				return;
			}
		}
		
		
		if(this->holdObject){
			
			
			if(__NEAR == this->direction.z){
				
				walkAnimation = walkFrontHoldStr;	
			}
			else{
				
				walkAnimation = walkBackHoldStr;
			}
		}
		else{
			
			if(__FAR == this->direction.z){
				
				walkAnimation = walkBackStr;
			}

		}

		// if it is only moving over the x axis and walking animation is no playing
		if(!AnimatedSprite_isPlayingFunction((AnimatedSprite)this->sprite, this->actorDefinition->animationDescription, walkAnimation)){
			
			// must play walk animation of not playing it
			Actor_playAnimation((Actor)this, walkAnimation);
		}

		// change state
		StateMachine_swapState(this->stateMachine, (State)MarioOnBridge_getInstance());
		
		this->movementState.z = __ACTIVE;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// update movement over the bridge
void Mario_moveOnBridge(Mario this){


	this->position.z += FIX23_9_MULT(MARIO_DISPLACEMENT_ON_BRIDGE, ITOFIX23_9(this->direction.z));
	
	// if I'm holding something
	if(this->holdObject){
		
		// update object's position
		Mario_updateHoldObjectPosition(this);
	}
	
	this->movementState.z = __ACTIVE;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// set bridge
void Mario_setBridge(Mario this, Bridge bridge){
	
	this->bridge = bridge;
}



//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// determine which layer I'm
void Mario_determineLayer(Mario this){

	int i = 0;
	for(i = 0; i < TOTAL_GAME_LAYERS; i++){
		
		if((u16)this->position.z > gameLayers[i] - ITOFIX23_9(5) && (unsigned)this->position.z < gameLayers[i] + ITOFIX23_9(5)){
			
			this->layer = (int )i;
			return;
		}
			
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// clear the actionTime
void Mario_resetActionTime(Mario this){

	this->actionTime = Clock_getTime(_clock);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// goal reached
void Mario_win(Mario this){
	
	this->direction.z = __FAR;
	
	Actor_playAnimation((Actor)this, "WalkBack");
	
	Actor_startZMovement((Actor)this, __UNIFORMMOVE, FIX23_9_DIV(MARIO_VELOCITY_Z, ITOFIX23_9(4)), 0);
	
	this->actionTime = Clock_getTime(_inGameClock);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// goal reached
void Mario_moveOnWin(Mario this){
	
	u32 currentTime = Clock_getTime(_inGameClock);
	
	if(this->movementState.z){
	
		if(currentTime - this->actionTime > MARIO_WIN_DELAY || 
			FIX23_9TOI(this->position.z) >= Stage_getSize(Game_getStage(Game_getInstance())).z){
			
			// inform the game that I'm dead
			GameWorld_levelCleared(GameWorld_getInstance());
		}
		
		/*
		if(currentTime - this->actionTime < MARIO_WIN_DELAY / 2){
				
			Actor_move((Actor)this);
		}
		return;
		*/
	}
	
	Actor_move((Actor)this);
}