/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */



/*---------------------------------INCLUDES--------------------------------*/
#include "Koopa.h"
#include "KoopaWalking.h"
#include "KoopaStunned.h"
#include "KoopaBouncing.h"
#include "../enemy/EnemyDead.h"
#include "../enemy/EnemyAttack.h"
#include "../mario/Mario.h"


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											 CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


// Koopa.c

// A Koopa!
__CLASS_DEFINITION(Koopa, Enemy); 

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */




/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// always call these to macros next to each other
__CLASS_NEW_DEFINITION(Koopa, __PARAMETERS(ActorDefinition* characterDefinition, int ID))
__CLASS_NEW_END(Koopa, __ARGUMENTS(this, characterDefinition, ID));

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's conctructor
void Koopa_constructor(Koopa this, ActorDefinition* characterDefinition, int ID){
	
	__CONSTRUCT_BASE(Enemy, __ARGUMENTS(this, characterDefinition, ID));

	// initialize me as walking
	StateMachine_swapState(this->stateMachine, (State)KoopaWalking_getInstance());
	
	// register a shape for collision detection
	Koopa_registerShape(this);

}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's conctructor
void Koopa_destructor(Koopa this){

	// delete the super object
	__DEALLOCATE(Enemy);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// register a shape with the collision detection system
void Koopa_registerShape(Koopa this){

	// register a shape for collision detection
	this->shape = CollisionManager_registerShape(CollisionManager_getInstance(), (InGameEntity)this, kRect, this->actorDefinition->inGameEntityDefinition.deep);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// process a collision
u8 Koopa_processCollision(Koopa this, Telegram telegram){
	
	int message = Telegram_getMessage(telegram);
	InGameEntity inGameEntity = (InGameEntity) Telegram_getExtraInfo(telegram);
	
	switch(message){


		case kCollisionXY:
		case kCollisionX:
		case kCollisionY:
		case kCollisionZ:			

			
			switch(InGameEntity_getInGameType(inGameEntity)){
			
				case kMario:

					// tell mario to take a hit
					// first ask if he is not hitting me
					if(Mario_isHitByEnemy((Mario)inGameEntity, (Enemy)this, message)){
						
						// ok, i hit him
						Mario_takeHit((Mario)inGameEntity, this->position);

						// now I attack
						Koopa_attack(this);
					}
					break;
					
				case kBridge:
					break;
					
				case kFloor:
				case kSolid:			

					
					return Koopa_checkIfTurn(this, inGameEntity, message);
					break;
				
					
				case kPiranhaPlant:
				case kKoopa:
					
					// if I'm bouncing... 
					if(StateMachine_isInState(this->stateMachine, (State) KoopaBouncing_getInstance())){
						
						// kill it
						__VIRTUAL_CALL(void, Enemy, die, (Enemy) inGameEntity);
						return false;
					}

					// just check if turn around
					return Koopa_checkIfTurn(this, inGameEntity, message);
					break;
			}
			
			
			break;
	}

	return false;

}




//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// update movement
void Koopa_walk(Koopa this){

	// move the actor
	Actor_move((Actor)this);
	
	// check if must turn because I'm on the edge of the object below
	if(Enemy_onPlatformEdge((Enemy)this)){

		// must change direction
		Actor_moveOpositeDirecion((Actor)this, __XAXIS);
		
		this->position.x += ITOFIX23_9(this->direction.x << 1);

		// play animation
		Actor_playAnimation((Actor)this, "Turn");
			
		// and stop moving
		Actor_stopXMovement((Actor)this);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// update movement
void Koopa_slide(Koopa this){

	// move the actor
	Actor_move((Actor)this);

}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// tell me I've been hit
void Koopa_takeHit(Koopa this, int axis, s8 direction){
	
	// if I'm idle
	if(StateMachine_isInState(this->stateMachine,(State)KoopaWalking_getInstance())){							
	
		// must be stunned now
		StateMachine_swapState(this->stateMachine, (State)KoopaStunned_getInstance());
	
		// record when I've been hit
		this->actionTime = Clock_getTime(_inGameClock);
		
		// play animation
		Actor_playAnimation((Actor)this, "Stunned");
	}
	else{		
		
		// I'm not moving, I assume I'm stunned 
		if (!Actor_isMoving((Actor)this)) {

			// must bounce
			Koopa_beThrown(this, axis, direction);
		}
		else{
			
			// otherwise, I assume I'm bouncing
			
			// record a net hit time
			this->actionTime = Clock_getTime(_inGameClock);
			
			// go again to stunned state
			StateMachine_swapState(this->stateMachine, (State)KoopaStunned_getInstance());
		}
	}

	// this forces gap calculation
	this->previousDirection.x = !this->direction.x;
	
	// make sure new gap is calculated now 
	// for correct collision detections
	Koopa_setGap((Koopa)this);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// go back to normal
void Koopa_recover(Koopa this){

	// play animation
	Actor_playAnimation((Actor)this, "Walk");
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// begin bouncing in a given direction
void Koopa_bounce(Koopa this, s8 direction){

	// set direction
	this->direction.x = direction;

	// begin movement
	Actor_stopMovement((Actor)this);
	Actor_startXMovement((Actor)this, __UNIFORMMOVE, KOOPA_BOUNCE_VELOCITY_X, KOOPA_ACCELERATION_X);

	// walk again
	StateMachine_swapState(this->stateMachine, (State)KoopaBouncing_getInstance());

	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// retrieve graphical gap
void Koopa_setGap(Koopa this){

	this->gap = this->inGameEntityDefinition->gap;
	
	// if I'm walking, return ROM gap
	if(!StateMachine_isInState(this->stateMachine,(State)KoopaWalking_getInstance())){

		this->gap.left = 4;
		this->gap.right = 4;
		
		this->gap.up = 12;
		this->gap.down = 1;
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Koopa_turnComplete(Koopa this){

	// start to move
	Actor_startXMovement((Actor)this, __UNIFORMMOVE, KOOPA_VELOCITY_X, KOOPA_ACCELERATION_X);
	
	// play walking animation 
	Actor_playAnimation((Actor)this, "Walk");
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// attack logic
void Koopa_attack(Koopa this){
	
	// if I'm walking
	if(StateMachine_isInState(this->stateMachine, (State)KoopaWalking_getInstance())){
		
		// play attack animation
		Actor_playAnimation((Actor)this, "Attack");
	}
	
	// change state
	StateMachine_pushState(this->stateMachine, (State)EnemyAttack_getInstance());

	// record a net hit time
	this->actionTime = Clock_getTime(_inGameClock);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// check if turn around
int Koopa_checkIfTurn(Koopa this, InGameEntity inGameEntity, int message){
	
	// check collision over each axis
	if(this->movementState.z){
	
		// align to colliding object
		Actor_alignZ((Actor)this, (InGameEntity)inGameEntity, 1);
		
		// must change direction
		Actor_moveOpositeDirecion((Actor)this, __ZAXIS);		
		return true;
	}
	
	// check if must turn around
	if(this->movementState.x && (kCollisionX == message || kCollisionXY == message)){
		
		// aling to colliding object
		Actor_alignX((Actor)this, (InGameEntity)inGameEntity, 1);
		
		// must change direction
		Actor_moveOpositeDirecion((Actor)this, __XAXIS);
		
		// if I'm walking... play turn animation
		if(StateMachine_isInState(this->stateMachine,(State)KoopaWalking_getInstance())){
			
			Actor_playAnimation((Actor)this, "Turn");
			
			// and stop moving
			Actor_stopXMovement((Actor)this);
		}
		
		return true;
	}
	
	
	// if was falling
	if(this->movementState.y && (kCollisionY == message || kCollisionXY == message)){
	
		// aling to below object
		Actor_alignY((Actor)this, (InGameEntity)inGameEntity, 1);
		
		// finally stop the movement
		Actor_stopYMovement((Actor)this);

		// save object below me
		this->objectBelow = inGameEntity;
		
		return true;
	}
	
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// die 
void Koopa_die(Koopa this){
	
	// if I'm walking
	if(StateMachine_isInState(this->stateMachine, (State)KoopaWalking_getInstance())){
	
		// play animation
		Actor_playAnimation((Actor)this, "Attack");
	}

	// must unregister the shape for collision detections
	Shape_setActive(this->shape, false);

	// now change state to dead
	StateMachine_swapState(this->stateMachine, (State)EnemyDead_getInstance());
	
	// set state
	this->inGameState = kDead;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// shake it
void Koopa_shake(Koopa this){

	if(__RIGHT == this->direction.x){

		// displace me over x
		this->position.x -=2;
		
		// change direction
		this->direction.x = __LEFT;
	}
	else{
		
		// displace me over x
		this->position.x +=2;
		
		// change direction
		this->direction.x = __RIGHT;
	}
}
	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// be thrown over an axis and a direction
void Koopa_beThrown(Koopa this, int axis, s8 direction){

	Actor_stopMovement((Actor)this);
	// walk again
	StateMachine_swapState(this->stateMachine, (State)KoopaBouncing_getInstance());

	if(__XAXIS == axis){

		this->direction.x = direction;
		// begin movement
		
		Actor_startXMovement((Actor)this, __UNIFORMMOVE, KOOPA_BOUNCE_VELOCITY_X, KOOPA_ACCELERATION_X);

	}
	else{

		this->direction.z = direction;
		
		// begin movement
		Actor_startZMovement((Actor)this, __UNIFORMMOVE, KOOPA_BOUNCE_VELOCITY_Z, KOOPA_ACCELERATION_Z);
	}
	
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// tell whether it is ready to attack mario
int Koopa_canAttack(Koopa this){
	
	return !StateMachine_isInState(this->stateMachine,	(State)KoopaStunned_getInstance());
}
