/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												INCLUDES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#include "MarioIdle.h"
#include "MarioMoving.h"
#include "Mario.h"

#ifdef __DEBUG

#include "../../worlds/GameWorld.h"
#endif

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


// class's constructor
static void MarioIdle_constructor(MarioIdle this);

// class's destructor
static void MarioIdle_destructor(MarioIdle this);

// state's enter
static void MarioIdle_enter(MarioIdle this, void* owner);

// state's execute
static void MarioIdle_execute(MarioIdle this, void* owner);

// state's exit
static void MarioIdle_exit(MarioIdle this, void* owner);

// state's pause
static void MarioIdle_pause(MarioIdle this, void* owner){}

// state's resume
static void MarioIdle_resume(MarioIdle this, void* owner){}

// state's on message
static u16 MarioIdle_handleMessage(MarioIdle this, void* owner, Telegram telegram);



/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

__CLASS_DEFINITION(MarioIdle, State);

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
__SINGLETON(MarioIdle);
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's constructor
static void MarioIdle_constructor(MarioIdle this){

	__STATE_CONSTRUCT(MarioIdle, State);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
static void MarioIdle_destructor(MarioIdle this){
	
	// allow another creation of this state
	__SINGLETON_DESTRUCT;	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// state's enter
static void MarioIdle_enter(MarioIdle this, void* owner){
	
	Mario_resetActionTime((Mario)owner);
	
	if(Mario_isMovingOverZ((Mario)owner)){
		
		if(__NEAR == InGameEntity_getDirection((InGameEntity)owner).z){

			// show animation
			Actor_playAnimation((Actor)owner, "Front");
		}
		else{
			
			// show animation
			Actor_playAnimation((Actor)owner, "Back");
			
		}
	}
	else{
	
		// show animation
		Actor_playAnimation((Actor)owner, "Idle");
	}
	
	Mario_resetActionTime((Mario)owner);

	Mario_setGap((Mario)owner);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// state's execute
static void MarioIdle_execute(MarioIdle this, void* owner){
	

	
	if(!Mario_isMovingOverZ((Mario)owner)){
		
		// if up key pressed
		if(!((K_LU | K_LD) & vbKeyPressed())){
	
			// make mario to look to the side
			Mario_lookSide((Mario)owner);
		}
	}

	if(!(K_B & vbKeyPressed())){
		
		// check if must thrown an object
		Mario_throwObject((Mario)owner);
	}
	
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// state's exit 
static void MarioIdle_exit(MarioIdle this, void* owner){
	
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// state's on message
static u16 MarioIdle_handleMessage(MarioIdle this, void* owner, Telegram telegram){

	switch(Telegram_getMessage(telegram)){
	
		case kNoObjectBelow:

			// begin to fall
			Actor_fall((Actor)owner);
			
			// change state
			StateMachine_swapState(Actor_getStateMachine((Actor) owner), (State)MarioMoving_getInstance());
			return true;
			break;
	
		case kCollisionX:
		case kCollisionY:
		case kCollisionZ:

			// process the collision
			return Mario_processCollision((Mario)owner, telegram);
			break;
			
		case kKeyPressed:
			
#ifdef __DEBUG
			if(kPlaying != GameWorld_getMode(GameWorld_getInstance())){
				
				return true;
			}
#endif
			
			if((K_LL | K_LR | K_A) & vbKeyPressed()){
				
					// set direction
					Direction direction = InGameEntity_getDirection((InGameEntity)owner);
					
					u16 currentKey = vbKeyPressed();
					
					if((K_LR & currentKey) && (K_LL & currentKey)){
						return true;
					}

					if(K_LR & currentKey){
						
						direction. x = __RIGHT;
					}
					else
					{
						if(K_LL & currentKey){
						
							direction. x = __LEFT;
						}
					}		
					
					InGameEntity_setDirection((InGameEntity)owner, direction);
				
					// start movement
					StateMachine_swapState(Actor_getStateMachine((Actor) owner),(State) MarioMoving_getInstance());
					
					if(K_A & currentKey){
						
						Mario_jump((Mario)owner);
					}
					return true;
					break;
					
			}
			
			// if up key pressed
			if(K_LU & vbKeyPressed()){

				// if already not playing
				if(!Actor_isPlayingAnimation((Actor)owner, "Back")){

					// make mario to look to the player
					Mario_lookBack((Mario)owner);
				}

				// check if there is a bridge
				Mario_checkIfBridge((Mario)owner, K_LU);

				// check if there is a bridge
				Mario_checkIfBridge((Mario)owner, K_LU);
				
				return true;
			}
			
			// if up key pressed
			if(K_LD & vbKeyPressed()){

				// if already not playing
				if(!Actor_isPlayingAnimation((Actor)owner, "Front")){

					// make mario to look away the player
					Mario_lookFront((Mario)owner);
				}

				// check if there is a bridge
				Mario_checkIfBridge((Mario)owner, K_LD);

				// check if there is a bridge
				Mario_checkIfBridge((Mario)owner, K_LD);

				return true;
			}	

			{
				static int alpha = 0;
				// if up key pressed
				if(K_LT & vbKeyPressed()){
					
					alpha--;
					if(alpha < 0){
						alpha = 511;
					}
					Sprite_rotate(Entity_getSprite((Entity)owner), alpha);
					
				}
				
				// if up key pressed
				if(K_RT & vbKeyPressed()){
					alpha++;
					if(alpha > 511){
						alpha = 0;
					}				
					Sprite_rotate(Entity_getSprite((Entity)owner), alpha);
					
				}
			}
			break;
	}
	return false;
}

