/* VbJaEngine: bitmap graphics engine for the Nintendo Virtual Boy 
 * 
 * Copyright (C) 2007 Jorge Eremiev
 * jorgech3@gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												INCLUDES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

#include "GameWorld.h"
#include "InitFocusScreen.h"
#include "VbJaeScreen.h"
#include "TitleScreen.h"
#include <objects.h>
#include <macros.h>


/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												PROTOTYPES
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

static void GameWorld_destructor(GameWorld this);

// class's constructor
static void GameWorld_constructor(GameWorld this);

// state's enter
static void GameWorld_enter(GameWorld this, void* owner);

// state's execute
static void GameWorld_execute(GameWorld this, void* owner);

// state's enter
static void GameWorld_exit(GameWorld this, void* owner);

// state's execute
static void GameWorld_pause(GameWorld this, void* owner){}

// state's execute
static void GameWorld_resume(GameWorld this, void* owner){}

// state's on message
static u16 GameWorld_handleMessage(GameWorld this, void* owner, Telegram telegram);

// move view point
static void GameWorld_movePerspective(GameWorld this);

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 											CLASS'S DEFINITION
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */


	
__CLASS_DEFINITION(GameWorld, State);

/* ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * 												CLASS'S METHODS
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 * ---------------------------------------------------------------------------------------------------------
 */

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// it's a singleton
__SINGLETON_DYNAMIC(GameWorld);



//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's constructor
static void GameWorld_constructor(GameWorld this){
		
	__STATE_CONSTRUCT(GameWorld, State);
	
	this->marioIsDead = false;
	
	this->levelCleared = false;
	
	this->stage = Game_getStage(Game_getInstance());
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class's destructor
static void GameWorld_destructor(GameWorld this){
	
	// allow another creation of this state
	__SINGLETON_DESTRUCT;	
	
	// destroy the super object
	__DEALLOCATE(State);	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// state's enter
static void GameWorld_enter(GameWorld this, void* owner){
	
	_asciiChar = (const u16*)ASCII_CH;
	
	// make a fade out
	//Game_FXFadeOut(Game_getInstance(),FADEDELAY);
	Game_FXFadeOut(Game_getInstance(),10);
	
	//clear char and bgmap memory
	vbClearScreen();
	
	// reset the engine state
	Game_reset(Game_getInstance(), true);
	
	
	//load world entities	
	Stage_load(this->stage, &WORLD_0_0_0);
	
	
	//call the render to setup world entries
	Game_render(Game_getInstance());
	 
	// reset the global clock
	//Clock_reset(Game_getClock(Game_getInstance()));
	
	Clock_reset(_inGameClock);
	Clock_start(_inGameClock);
	
	// turn back the background
	VIP_REGS[BKCOL] = 0x00;

	// make a fade in
	Game_FXFadeIn(Game_getInstance(), FADEDELAY);
	
	// mario is alive!
	this->marioIsDead = false;
	
	this->levelCleared = false;
	
	// playing by default
	this->mode = kPlaying;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// state's execute
static void GameWorld_execute(GameWorld this, void* owner){

	// if mario is dead
	if(this->marioIsDead){
		
		// force exit and enter to next state
		Game_setState(Game_getInstance(), (State)TitleScreen_getInstance());
			
		return;
	}

	// if mario is dead
	if(this->levelCleared){
		
		// force exit and enter to next state
		Game_setState(Game_getInstance(), (State)VbJaeScreen_getInstance());

		return;
	}

#ifdef __DEBUG
	if(kPlaying == this->mode){
#endif	
		// center the screen
		Stage_centerScreen(this->stage);

#ifdef __DEBUG		
	}
#endif
	
	// if the screen is moving
	if(*((u8*)_screenMovementState)){

		static int turn = __TARGET_FPS >> 4;
			
		if(turn--){

			// unload not visible objects	
			Stage_unloadEntities(this->stage);
		}
		else{

			// load visible objects	
			Stage_loadEntities(this->stage);
			
			turn = __TARGET_FPS >> 4;
		}
			
	}
	else{
		
		static int delay = __TARGET_FPS >> 2;
		
		if(!delay--){

			// unload objects which are moving outside the screen	
			Stage_unloadEntities(this->stage);
			
			delay = __TARGET_FPS >> 2;
		}
	}

	if(kPaused != this->mode){
		
		// update game world objects
		Stage_update(this->stage);
	}
		
	// now sort the world layers
	Stage_checkLayers();
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// state's exit 
static void GameWorld_exit(GameWorld this, void* owner){
	
	// destroy the state
	__DELETE(this);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// state's on message
static u16 GameWorld_handleMessage(GameWorld this, void* owner, Telegram telegram){

	
	// process message
	switch(Telegram_getMessage(telegram)){
	
		case kKeyPressed:	
			
			//if start button pressed
			if(vbNewKeyPressed(K_STA)){
				
				if(kPaused == this->mode){					
					
					this->mode = kPlaying;
					vbjPrintText("     ", 22, 13);
					Clock_pause(_inGameClock, false);
					
					VIP_REGS[GPLT0] = __GPLT0VALUE;	/* Set all eight palettes to: 11100100 */
					VIP_REGS[GPLT1] = __GPLT1VALUE;	/* (i.e. "Normal" dark to light progression.) */
					VIP_REGS[GPLT2] = __GPLT2VALUE;
					VIP_REGS[GPLT3] = __GPLT3VALUE;
					
				}
				else{
					
					vbjPrintText("                  ", 0, 0);
					this->mode = kPaused;
					vbjPrintText("Pause", 22, 13);
					Clock_pause(_inGameClock, true);
					
					VIP_REGS[GPLT0] = 0xA0;	/* Set all eight palettes to: 11100100 */
					VIP_REGS[GPLT1] = 0x90;	/* (i.e. "Normal" dark to light progression.) */
					
				}
				return true;
			}

			if(kPaused == this->mode){
				
				return true;
			}
			

#ifdef __DEBUG
			//if select button pressed
			if(vbNewKeyPressed(K_SEL)){
				this->mode++;
				if(this->mode > GAMEWORLD_MODES - 1){
					this->mode = 1;
				}
				
				if(kMovingScreen == this->mode){
					
					this->lastTime = Clock_getTime(_inGameClock);
					vbjPrintText("Moving screen     ", 0, 0);
				}
				
				if(kMovingPerspective == this->mode){
					
					this->lastTime = Clock_getTime(_inGameClock);
					vbjPrintText("Moving perspective", 0, 0);
				}
				
				if(kPlaying == this->mode){
					
					this->lastTime = Clock_getTime(_inGameClock);
					vbjPrintText("                  ", 0, 0);
				}				
			}
			
			if(kMovingScreen == this->mode){
				
				if((K_LU | K_LD | K_LL | K_LR) & vbKeyPressed()){

					GameWorld_moveScreen(this);

				}
			}
			
			if(kMovingPerspective == this->mode){
				
				if(vbKeyPressed()){

					GameWorld_movePerspective(this);
				}
			}			
#endif 
			break;
	}

	return false;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// mario is dead
void GameWorld_marioIsDead(GameWorld this){

	// now he's dead
	this->marioIsDead = true;
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// level completed
void GameWorld_levelCleared(GameWorld this){

	// now he's dead
	this->levelCleared = true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// move the screen
void GameWorld_moveScreen(GameWorld this){

	u32 currentTime = Clock_getTime(_inGameClock);
	int displacement = 200 * (currentTime - this->lastTime) / 1000;

	if(currentTime - this->lastTime > 200){
		
		this->lastTime = Clock_getTime(_inGameClock);;
		displacement = 4;
	}
	this->lastTime = currentTime;
	
	displacement = ITOFIX23_9(displacement);
	
	
	if(vbKeyPressed() & K_LL){
		
		if(_screenPosition->x - displacement > 0){
			
			_screenPosition->x -= displacement;
			
			_screenMovementState->x = __ACTIVE;
		}
	}
	
	if(vbKeyPressed() & K_LR){
		
		//if(_screenPosition.x > displacement > 0){
			
			_screenPosition->x += displacement;
			
			_screenMovementState->x = __ACTIVE;
			
		//}
	}	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// move view point
static void GameWorld_movePerspective(GameWorld this){

	switch(vbKeyPressed()){
	
		case K_LL:
			
			if(_optical->horizontalViewPointCenter > 0){
				
				_optical->horizontalViewPointCenter--;
			}
			else{
				_optical->horizontalViewPointCenter=0;
			}
			
			break;
			
		case K_LR:
			
			if(_optical->horizontalViewPointCenter < 384){
				
				_optical->horizontalViewPointCenter++;
			}
			else{
				_optical->horizontalViewPointCenter=384;
			}

			break;
			
		case K_LU:
			
			if(_optical->verticalViewPointCenter > 0){
				
				_optical->verticalViewPointCenter--;
			}
			else{
				_optical->verticalViewPointCenter=0;
			}

			break;
			
		case K_LD:
			
			if(_optical->verticalViewPointCenter < 224){
				
				_optical->verticalViewPointCenter++;
			}
			else{
				_optical->verticalViewPointCenter=224;
			}

			break;
			
		case K_RL:
			
			if(_optical->baseDistance > 10){
				
				_optical->baseDistance -=10;
			}
			else{
				_optical->baseDistance = 1;
			}
	
			break;
			
		case K_RR:
			
			if(_optical->baseDistance < __BASEFACTOR +__BASEFACTOR/2){
				
				_optical->baseDistance += 10;
			}
			else{
				_optical->baseDistance=__BASEFACTOR+__BASEFACTOR/2;
			}
	
			break;

			/*
		case K_RU:
			
			if(_optical->maximunViewDistance > __MAXVIEWDISTANCE/2){
				
				_optical->maximunViewDistance--;
			}
			else{
				
				_optical->maximunViewDistance =__MAXVIEWDISTANCE/2;
			}

			
			break;
			
		case K_RD:
			
			if(_optical->maximunViewDistance < __MAXVIEWDISTANCE+__MAXVIEWDISTANCE/2){
				
				_optical->maximunViewDistance++;
			}
			else{
				_optical->maximunViewDistance = __MAXVIEWDISTANCE+__MAXVIEWDISTANCE/2;
			}

			break;
			*/
		}
		
	
	_screenMovementState->x = _screenMovementState->y = _screenMovementState->z = __ACTIVE; 

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// get working mode
int GameWorld_getMode(GameWorld this){
	
	return this->mode;
}
